#!/usr/bin/env python3

# pylint: disable=missing-docstring

import ast
import inspect
import pathlib
import pickle
import re
import textwrap
import unittest
from typing import List, Dict, Optional, Final, Sequence

from icontract import require

from aas_core_codegen import intermediate
from aas_core_codegen.common import Identifier

# noinspection PyProtectedMember
from aas_core_codegen.intermediate import _types as intermediate_types

import tests.common


class TestPickle(unittest.TestCase):
    def test_enumeration(self) -> None:
        source = textwrap.dedent(
            """\
            from enum import Enum

            class Some_enum(Enum):
                Literal1 = "literal_1"
                Literal2 = "literal_2"

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None and symbol_table is not None

        some_enum = symbol_table.must_find_enumeration(Identifier("Some_enum"))

        # NOTE (mristin):
        # We first test that ID sets are working before pickling.
        original_id_set = some_enum.literal_id_set
        self.assertEqual(len(original_id_set), 2)
        self.assertIn(id(some_enum.literals[0]), original_id_set)
        self.assertIn(id(some_enum.literals[1]), original_id_set)

        pickled_data = pickle.dumps(some_enum)
        unpickled_enum = pickle.loads(pickled_data)

        # NOTE (mristin):
        # We now test that the unpickled ID sets are also working.
        self.assertEqual(unpickled_enum.name, "Some_enum")
        self.assertEqual(len(unpickled_enum.literals), 2)
        self.assertEqual(unpickled_enum.literals[0].name, "Literal1")
        self.assertEqual(unpickled_enum.literals[1].name, "Literal2")

        new_id_set = unpickled_enum.literal_id_set
        self.assertEqual(len(new_id_set), 2)
        self.assertIn(id(unpickled_enum.literals[0]), new_id_set)
        self.assertIn(id(unpickled_enum.literals[1]), new_id_set)

        # NOTE (mristin):
        # The ID sets should be different since objects have new IDs after unpickling.
        self.assertNotEqual(original_id_set, new_id_set)

    def test_abstract_class(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Some_abstract_class:
                some_property: int

                @require(lambda some_property: some_property > 0)
                def __init__(self, some_property: int) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None and symbol_table is not None

        some_abstract_class = symbol_table.must_find_abstract_class(
            Identifier("Some_abstract_class")
        )

        pickled_data = pickle.dumps(some_abstract_class)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.AbstractClass)

        self.assertEqual(unpickled.name, "Some_abstract_class")
        self.assertEqual(len(unpickled.properties), 1)
        self.assertEqual(unpickled.properties[0].name, "some_property")

        # Test that ID sets are rebuilt correctly
        self.assertIn(id(unpickled.properties[0]), unpickled.property_id_set)

    def test_argument(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                def some_method(self, some_arg: int) -> None:
                    pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None and symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_method = some_class.methods[0]
        some_arg = some_method.arguments[0]

        pickled_data = pickle.dumps(some_arg)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.Argument)

        self.assertEqual(unpickled.name, "some_arg")

    def test_class(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                some_property: int
                
                def __init__(
                        self,
                        some_property: int
                ) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_class(Identifier("Some_class"))

        pickled_data = pickle.dumps(some_class)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.ConcreteClass)

        self.assertEqual(unpickled.name, "Some_class")

        self.assertIn(id(unpickled.properties[0]), unpickled.property_id_set)

    def test_concrete_class(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_concrete_class:
                some_property: int

                def __init__(
                        self,
                        some_property: int
                ) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_concrete_class = symbol_table.must_find_concrete_class(
            Identifier("Some_concrete_class")
        )

        pickled_data = pickle.dumps(some_concrete_class)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.ConcreteClass)

        self.assertEqual(unpickled.name, "Some_concrete_class")
        self.assertEqual(len(unpickled.properties), 1)

        self.assertIn(id(unpickled.properties[0]), unpickled.property_id_set)

    def test_constant_primitive(self) -> None:
        source = textwrap.dedent(
            """\
            Some_constant: str = constant_str(
                value="some_value",
            )

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_constant = symbol_table.must_find_constant_primitive(
            Identifier("Some_constant")
        )

        pickled_data = pickle.dumps(some_constant)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.ConstantPrimitive)

        self.assertEqual(unpickled.name, "Some_constant")
        self.assertEqual(unpickled.value, "some_value")

    def test_constant_set_of_primitives(self) -> None:
        source = textwrap.dedent(
            """\
            from aas_core_meta.marker import (
                constant_set
            )

            Some_constant_set: Set[str] = constant_set(
                values=["value1", "value2"]
            )

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_constant_set = symbol_table.must_find_constant_set_of_primitives(
            Identifier("Some_constant_set")
        )

        pickled_data = pickle.dumps(some_constant_set)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.ConstantSetOfPrimitives)

        self.assertEqual(unpickled.name, "Some_constant_set")
        self.assertEqual(len(unpickled.literals), 2)

    def test_constrained_primitive(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: len(self) > 0, "Non-empty")
            class Some_constrained_primitive(str):
                pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_constrained_primitive = symbol_table.must_find_constrained_primitive(
            Identifier("Some_constrained_primitive")
        )

        pickled_data = pickle.dumps(some_constrained_primitive)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.ConstrainedPrimitive)

        self.assertEqual(unpickled.name, "Some_constrained_primitive")

        self.assertIn(id(unpickled.invariants[0]), unpickled.invariant_id_set)

    def test_constructor(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                some_property: int

                def __init__(self, some_property: int) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        constructor = some_class.constructor

        pickled_data = pickle.dumps(constructor)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.Constructor)

        self.assertEqual(len(unpickled.arguments), 1)

    def test_contract(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                @require(lambda self: True)
                @ensure(lambda result: True)
                def some_method(self) -> bool:
                    return True

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_method = some_class.methods[0]
        contract = some_method.contracts.preconditions[0]

        pickled_data = pickle.dumps(contract)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.Contract)

        self.assertIsNotNone(unpickled.body)

    def test_contracts(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                @require(lambda self: True)
                def some_method(self) -> None:
                    pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_method = some_class.methods[0]
        contracts = some_method.contracts

        pickled_data = pickle.dumps(contracts)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.Contracts)

        self.assertEqual(len(unpickled.preconditions), 1)

    def test_default_primitive(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                some_property: str
            
                def __init__(self, some_property: str = 'some_default') -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))

        arg = some_class.constructor.arguments_by_name[Identifier("some_property")]

        default = arg.default

        pickled_data = pickle.dumps(default)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.DefaultPrimitive)

    def test_default_enumeration_literal(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_enum(Enum):
                Literal1 = "literal-1"
                Literal2 = "literal-2"
            
            class Some_class:
                some_property: Some_enum

                def __init__(
                        self, 
                        some_property: Some_enum = Some_enum.Literal1
                ) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))

        arg = some_class.constructor.arguments_by_name[Identifier("some_property")]

        default = arg.default

        pickled_data = pickle.dumps(default)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.DefaultEnumerationLiteral)

    def test_description_of_constant(self) -> None:
        source = textwrap.dedent(
            """\
            Some_constant: str = constant_str(
                value="some_value",
                description="This is some constant."
            )

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_constant = symbol_table.must_find_constant_primitive(
            Identifier("Some_constant")
        )
        description = some_constant.description
        assert description is not None

        pickled_data = pickle.dumps(description)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.DescriptionOfConstant), str(
            type(unpickled)
        )

        self.assertEqual(unpickled.summary[0], "This is some constant.")

    def test_description_of_enumeration_literal(self) -> None:
        source = textwrap.dedent(
            """\
            from enum import Enum

            class Some_enum(Enum):
                literal1 = "value1"
                \"\"\"Describe the literal.\"\"\"

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_enum = symbol_table.must_find_enumeration(Identifier("Some_enum"))
        literal = some_enum.literals[0]
        description = literal.description

        pickled_data = pickle.dumps(description)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.DescriptionOfEnumerationLiteral)

        self.assertEqual(unpickled.summary[0], "Describe the literal.")

    def test_description_of_meta_model(self) -> None:
        source = textwrap.dedent(
            """\
            \"\"\"Meta-model description.\"\"\"

            class Some_class:
                pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        meta_model = symbol_table.meta_model
        description = meta_model.description

        pickled_data = pickle.dumps(description)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.DescriptionOfMetaModel)

        self.assertEqual(unpickled.summary[0], "Meta-model description.")

    def test_description_of_our_type(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                \"\"\"Describe the class.\"\"\"
                pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        description = some_class.description

        pickled_data = pickle.dumps(description)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.DescriptionOfOurType)

        self.assertEqual(unpickled.summary[0], "Describe the class.")

    def test_description_of_property(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                some_property: int
                \"\"\"Describe the property.\"\"\"
                
                def __init__(
                        self,
                        some_property: int
                ) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_property = some_class.properties[0]
        description = some_property.description

        pickled_data = pickle.dumps(description)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.DescriptionOfProperty)

        self.assertEqual(unpickled.summary[0], "Describe the property.")

    def test_description_of_signature(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                def some_method(self) -> None:
                    \"\"\"Describe the method.\"\"\"
                    pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_method = some_class.methods[0]
        description = some_method.description

        pickled_data = pickle.dumps(description)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.DescriptionOfSignature)

        self.assertEqual(unpickled.summary[0], "Describe the method.")

    def test_enumeration_literal(self) -> None:
        source = textwrap.dedent(
            """\
            from enum import Enum

            class Some_enum(Enum):
                literal1 = "value1"

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_enum = symbol_table.must_find_enumeration(Identifier("Some_enum"))
        literal = some_enum.literals[0]

        pickled_data = pickle.dumps(literal)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.EnumerationLiteral)

        self.assertEqual(unpickled.name, "literal1")
        self.assertEqual(unpickled.value, "value1")

    def test_implementation_specific_method(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                @implementation_specific
                def some_method(self) -> None:
                    pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_method = some_class.methods[0]

        pickled_data = pickle.dumps(some_method)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.ImplementationSpecificMethod)

        self.assertEqual(unpickled.name, "some_method")

    def test_implementation_specific_verification(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            @implementation_specific
            def some_verification(x: int) -> bool:
                pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        verification = symbol_table.verification_functions[0]

        pickled_data = pickle.dumps(verification)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.ImplementationSpecificVerification)

        self.assertEqual(unpickled.name, "some_verification")

    def test_interface(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Some_abstract_class:
                some_property: int
                
                def __init__(
                        self,
                        some_property: int
                ) -> None:
                    self.some_property = some_property
                
                def some_func(self) -> None:
                    pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_abstract_class = symbol_table.must_find_abstract_class(
            Identifier("Some_abstract_class")
        )

        interface = some_abstract_class.interface

        pickled_data = pickle.dumps(interface)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.Interface)

        self.assertEqual(unpickled.name, "Some_abstract_class")

        self.assertIn(id(unpickled.properties[0]), unpickled.property_id_set)

    def test_invariant(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(lambda self: self.some_property > 0, "Some property is positive")
            class Some_class:
                some_property: int                                

                def __init__(self, some_property: int) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        invariant = some_class.invariants[0]

        pickled_data = pickle.dumps(invariant)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.Invariant)

        self.assertIsNotNone(unpickled.body)

    def test_list_type_annotation(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                some_property: List[str]
                
                def __init__(
                        self,
                        some_property: List[str]
                ) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_property = some_class.properties[0]
        type_annotation = some_property.type_annotation

        pickled_data = pickle.dumps(type_annotation)
        unpickled = pickle.loads(pickled_data)

        self.assertIsInstance(unpickled, intermediate_types.ListTypeAnnotation)

    def test_meta_model(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        meta_model = symbol_table.meta_model

        pickled_data = pickle.dumps(meta_model)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate_types.MetaModel)

        self.assertEqual(unpickled.version, "dummy")

    def test_method(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                def some_method(self) -> None:
                    pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_method = some_class.methods[0]

        pickled_data = pickle.dumps(some_method)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.UnderstoodMethod)

        self.assertEqual(unpickled.name, "some_method")

    def test_optional_type_annotation(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                some_property: Optional[str]

                def __init__(
                    self,
                    some_property: Optional[str] = None
                ) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_property = some_class.properties[0]
        type_annotation = some_property.type_annotation

        pickled_data = pickle.dumps(type_annotation)
        unpickled = pickle.loads(pickled_data)

        self.assertIsInstance(unpickled, intermediate_types.OptionalTypeAnnotation)

    def test_our_type_annotation(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                pass

            class Some_another_class:
                some_property: Some_class
                
                def __init__(
                        self,
                        some_property: Some_class
                ) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_another_class = symbol_table.must_find_concrete_class(
            Identifier("Some_another_class")
        )
        some_property = some_another_class.properties[0]
        type_annotation = some_property.type_annotation

        pickled_data = pickle.dumps(type_annotation)
        unpickled = pickle.loads(pickled_data)

        self.assertIsInstance(unpickled, intermediate_types.OurTypeAnnotation)

    def test_pattern_verification(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def some_verification(x: str) -> bool:
                return match(r"^[a-z]+$", x) is not None

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        verification = symbol_table.verification_functions[0]

        pickled_data = pickle.dumps(verification)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.PatternVerification)

        self.assertEqual(unpickled.name, "some_verification")

    def test_primitive_set_literal(self) -> None:
        source = textwrap.dedent(
            """\
            from aas_core_meta.marker import (
                constant_set
            )

            Some_constant_set: Set[str] = constant_set(
                values=["value1", "value2"]
            )

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_constant_set = symbol_table.must_find_constant_set_of_primitives(
            Identifier("Some_constant_set")
        )
        literal = some_constant_set.literals[0]

        pickled_data = pickle.dumps(literal)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.PrimitiveSetLiteral)

        self.assertEqual(unpickled.value, "value1")

    def test_primitive_type(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                some_property: str
                
                def __init__(
                        self,
                        some_property: str
                ) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_property = some_class.properties[0]
        type_annotation = some_property.type_annotation
        assert isinstance(type_annotation, intermediate_types.PrimitiveTypeAnnotation)
        primitive_type = type_annotation.a_type

        pickled_data = pickle.dumps(primitive_type)
        unpickled = pickle.loads(pickled_data)

        self.assertEqual(unpickled, intermediate_types.PrimitiveType.STR)

    def test_primitive_type_annotation(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                some_property: str
                
                def __init__(
                        self,
                        some_property: str
                ) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_property = some_class.properties[0]
        type_annotation = some_property.type_annotation

        pickled_data = pickle.dumps(type_annotation)
        unpickled = pickle.loads(pickled_data)

        self.assertIsInstance(unpickled, intermediate_types.PrimitiveTypeAnnotation)

    def test_property(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                some_property: str
                
                def __init__(
                        self,
                        some_property: str
                ) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_property = some_class.properties[0]

        pickled_data = pickle.dumps(some_property)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.Property)

        self.assertEqual(unpickled.name, "some_property")

    def test_serialization(self) -> None:
        source = textwrap.dedent(
            """\
            @serialization(with_model_type=True)
            class Some_class:
                some_property: str
                
                def __init__(self, some_property: str) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        serialization = some_class.serialization

        pickled_data = pickle.dumps(serialization)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.Serialization)

        self.assertEqual(unpickled.with_model_type, True)

    def test_signature(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Some_abstract_class:
                some_property: int
            
                def __init__(self, some_property: int) -> None:
                    self.some_property = some_property
            
                def some_func(self) -> None:
                    pass
            
            
            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_abstract_class = symbol_table.must_find_abstract_class(
            Identifier("Some_abstract_class")
        )

        signature = some_abstract_class.interface.signatures[0]

        pickled_data = pickle.dumps(signature)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.Signature)

        self.assertEqual(unpickled.name, "some_func")

    def test_signature_like(self) -> None:
        source = textwrap.dedent(
            """\
            @abstract
            class Some_abstract_class:
                some_property: int

                def __init__(self, some_property: int) -> None:
                    self.some_property = some_property
            
                def some_func(self) -> None:
                    pass
            
            
            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        source = textwrap.dedent(
            """\
            @abstract
            class Some_abstract_class:
                some_property: int

                @require(lambda some_property: some_property > 0)
                def __init__(self, some_property: int) -> None:
                    self.some_property = some_property

                def some_func(self) -> None:
                    pass


            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_abstract_class = symbol_table.must_find_abstract_class(
            Identifier("Some_abstract_class")
        )

        signature_like: intermediate_types.SignatureLike = some_abstract_class.methods[
            0
        ]

        pickled_data = pickle.dumps(signature_like)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate_types.SignatureLike)

        self.assertEqual(unpickled.name, "some_func")

    def test_snapshot(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                some_property: int
                
                def __init__(self, some_property: int) -> None:
                    self.some_property = some_property
            
                @snapshot(lambda self: self.some_property + 1)
                def some_method(self) -> None:
                    pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_method = some_class.methods[0]
        snapshot = some_method.contracts.snapshots[0]

        pickled_data = pickle.dumps(snapshot)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.Snapshot)

        self.assertIsNotNone(unpickled.body)

    def test_summary_remarks_constraints_description(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                \"\"\"
                Summary line.

                Some remark.
                    
                :constraint A10: Soem constraint
                \"\"\"

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        description = some_class.description

        pickled_data = pickle.dumps(description)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.SummaryRemarksConstraintsDescription)

        self.assertEqual(unpickled.summary[0], "Summary line.")

    def test_summary_remarks_description(self) -> None:
        source = textwrap.dedent(
            """\
            from enum import Enum

            class Some_class:
                \"\"\"
                Summary line.

                Some remarks.

                Some description.
                \"\"\"

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_class(Identifier("Some_class"))
        description = some_class.description

        pickled_data = pickle.dumps(description)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.SummaryRemarksDescription), str(
            type(unpickled)
        )

        self.assertEqual(unpickled.summary[0], "Summary line.")

    def test_symbol_table(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        pickled_data = pickle.dumps(symbol_table)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.SymbolTable)

        self.assertEqual(len(unpickled.concrete_classes), 1)

    def test_transpilable_verification(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def some_verification(x: int) -> bool:
                return x > 0

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        verification = symbol_table.verification_functions[0]

        pickled_data = pickle.dumps(verification)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.TranspilableVerification)

        self.assertEqual(unpickled.name, "some_verification")

    def test_type_annotation(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                some_property: str
                
                def __init__(
                        self,
                        some_property: str
                ) -> None:
                    self.some_property = some_property

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_property = some_class.properties[0]
        type_annotation = some_property.type_annotation

        pickled_data = pickle.dumps(type_annotation)
        unpickled = pickle.loads(pickled_data)

        self.assertIsInstance(unpickled, intermediate_types.TypeAnnotation)

    def test_understood_method(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                def some_method(self) -> None:
                    pass

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_class = symbol_table.must_find_concrete_class(Identifier("Some_class"))
        some_method = some_class.methods[0]

        pickled_data = pickle.dumps(some_method)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.UnderstoodMethod)

        self.assertEqual(unpickled.name, "some_method")

    def test_verification(self) -> None:
        source = textwrap.dedent(
            """\
            @verification
            def some_verification(x: int) -> bool:
                return x > 0

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        verification = symbol_table.verification_functions[0]

        pickled_data = pickle.dumps(verification)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.Verification)

        self.assertEqual(unpickled.name, "some_verification")

    def test_constant_set_of_enumeration_literals(self) -> None:
        source = textwrap.dedent(
            """\
            from enum import Enum

            from aas_core_meta.marker import (
                constant_set
            )

            class Some_enum(Enum):
                Literal1 = "lit1"
                Literal2 = "lit2"

            Some_constant_set: Set[Some_enum] = constant_set(
                values=[
                    Some_enum.Literal1
                ]
            )

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        if error is not None:
            raise AssertionError(tests.common.most_underlying_messages(error))
        assert symbol_table is not None

        some_constant_set = symbol_table.must_find_constant_set_of_enumeration_literals(
            Identifier("Some_constant_set")
        )

        # Test before pickling
        original_id_set = some_constant_set.literal_id_set
        self.assertEqual(len(original_id_set), 1)
        self.assertIn(id(some_constant_set.literals[0]), original_id_set)

        # Pickle and unpickle
        pickled_data = pickle.dumps(some_constant_set)
        unpickled = pickle.loads(pickled_data)

        assert isinstance(unpickled, intermediate.ConstantSetOfEnumerationLiterals)

        # Test after unpickling
        self.assertEqual(unpickled.name, "Some_constant_set")
        self.assertEqual(len(unpickled.literals), 1)
        self.assertEqual(unpickled.literals[0].name, "Literal1")

        # Test that ID sets are rebuilt correctly
        new_id_set = unpickled.literal_id_set
        self.assertEqual(len(new_id_set), 1)
        self.assertIn(id(unpickled.literals[0]), new_id_set)


def _to_lower_snake_case(cls_name: str) -> str:
    """
    Convert a class name from CamelCase to snake_case.

    >>> _to_lower_snake_case("Enumeration")
    'enumeration'

    >>> _to_lower_snake_case("EnumerationLiteral")
    'enumeration_literal'

    >>> _to_lower_snake_case("ConstrainedPrimitive")
    'constrained_primitive'

    >>> _to_lower_snake_case("ConstructorArgumentOfClass")
    'constructor_argument_of_class'

    >>> _to_lower_snake_case("_ConstructorArgumentOfClass")
    'constructor_argument_of_class'
    """
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", cls_name.lstrip("_"))
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


class TestAssertions(unittest.TestCase):
    def test_coverage(self) -> None:
        """Assert that all classes from intermediate._types are covered in test methods."""
        all_classes = []  # type: List[str]
        for class_name, obj in inspect.getmembers(intermediate_types, inspect.isclass):
            if obj.__module__ == intermediate_types.__name__:
                all_classes.append(class_name)

        expected_test_prefixes = [
            f"test_{_to_lower_snake_case(cls_name)}" for cls_name in all_classes
        ]

        test_methods = []  # type: List[str]
        for method_name in dir(TestPickle):
            if method_name.startswith("test_") and callable(
                getattr(TestPickle, method_name)
            ):
                test_methods.append(method_name)

        missing_tests = []  # type: List[str]
        for expected_prefix in expected_test_prefixes:
            found = any(
                test_method.startswith(expected_prefix) for test_method in test_methods
            )
            if not found:
                missing_tests.append(expected_prefix)

        if len(missing_tests) > 0:
            missing_classes = []
            for prefix in missing_tests:
                assert prefix.startswith("test_") and len("test_") == 5
                snake_case_name = prefix[5:]

                parts = snake_case_name.split("_")
                original_name = "".join(word.capitalize() for word in parts)
                missing_classes.append(original_name)

            missing_tests_joined = ",\n".join(missing_tests)

            raise AssertionError(
                f"Missing one or more test methods "
                f"in {TestPickle.__name__} for {len(missing_tests)} class(es):\n"
                f"{missing_tests_joined}"
            )

    def test_getstate_setstate_consistency(self) -> None:
        class _PickableClass:
            """Represent a class which has pickle/unpickle methods."""

            getstate: Final[Optional[ast.FunctionDef]]
            setstate: Final[Optional[ast.FunctionDef]]
            id_set_properties: Final[Sequence[str]]

            @require(
                lambda getstate: not (getstate is not None)
                or getstate.name == "__getstate__"
            )
            @require(
                lambda setstate: not (setstate is not None)
                or setstate.name == "__setstate__"
            )
            def __init__(
                self,
                id_set_properties: Sequence[str],
                getstate: Optional[ast.FunctionDef] = None,
                setstate: Optional[ast.FunctionDef] = None,
            ) -> None:
                self.id_set_properties = id_set_properties
                self.getstate = getstate
                self.setstate = setstate

        types_file = pathlib.Path(intermediate_types.__file__)
        source_code = types_file.read_text(encoding="utf-8")

        tree = ast.parse(source_code)

        class_map = {}  # type: Dict[str, _PickableClass]

        for node in ast.walk(tree):
            if isinstance(node, ast.ClassDef):
                id_set_properties = []  # type: List[str]

                # NOTE (mristin):
                # We go through all type-annotated properties and pick all properties
                # with suffix ``*_id_set``. This will give us some guard rails to make
                # sure that we did not forget to pop/set these properties in
                # ``__getstate__`` and ``__setstate__``. However, there might be more
                # runtime properties which need checking. Eventually, though unlikely,
                # the suffix ``_id_set`` might be misleading, so then we have to add
                # white listing here.
                for item in node.body:
                    if isinstance(item, ast.AnnAssign):
                        if isinstance(item.target, ast.Name):
                            property_name = item.target.id
                            if property_name.endswith("_id_set"):
                                id_set_properties.append(property_name)

                getstate = None  # type: Optional[ast.FunctionDef]
                setstate = None  # type: Optional[ast.FunctionDef]

                for item in node.body:
                    if isinstance(item, ast.FunctionDef):
                        if item.name == "__getstate__":
                            getstate = item

                        elif item.name == "__setstate__":
                            setstate = item

                        else:
                            pass

                if getstate is not None or setstate is not None:
                    class_map[node.name] = _PickableClass(
                        id_set_properties=id_set_properties,
                        getstate=getstate,
                        setstate=setstate,
                    )

        for class_name, cls in class_map.items():
            if cls.getstate is None and cls.setstate is None:
                continue

            if cls.getstate is None and cls.setstate is not None:
                raise AssertionError(
                    f"Unexpected missing __getstate__ when __setstate__ is specified "
                    f"in the class {class_name!r}"
                )

            if cls.getstate is not None and cls.setstate is None:
                raise AssertionError(
                    f"Unexpected missing __setstate__ when __getstate__ is specified "
                    f"in the class {class_name!r}"
                )

            assert cls.getstate is not None and cls.setstate is not None

            # NOTE (mristin):
            # We extract popped attributes from ``__getstate__``.
            popped_attrs = []  # type: List[str]

            for node in ast.walk(cls.getstate):
                if (
                    isinstance(node, ast.Call)
                    and isinstance(node.func, ast.Attribute)
                    and node.func.attr == "pop"
                    and isinstance(node.func.value, ast.Name)
                    and node.func.value.id == "state"
                ):
                    if (
                        len(node.args) > 0
                        and isinstance(node.args[0], ast.Constant)
                        and isinstance(node.args[0].value, str)
                    ):
                        popped_attrs.append(node.args[0].value)

            setattr_attrs = []  # type: List[str]

            for node in ast.walk(cls.setstate):
                if (
                    isinstance(node, ast.Call)
                    and isinstance(node.func, ast.Name)
                    and node.func.id == "setattr"
                ):
                    # NOTE (mristin):
                    # The second argument is the name of the attribute.
                    if (
                        len(node.args) >= 2
                        and isinstance(node.args[0], ast.Name)
                        and node.args[0].id == "self"
                        and isinstance(node.args[1], ast.Constant)
                        and isinstance(node.args[1].value, str)
                    ):
                        setattr_attrs.append(node.args[1].value)

            if set(popped_attrs) != set(setattr_attrs):
                raise AssertionError(
                    f"In class {class_name!r}, there is a mismatch between "
                    f"popped attributes in __getstate__: {sorted(popped_attrs)}, "
                    f"and setattr attributes in __setstate__: {sorted(setattr_attrs)}"
                )

            # NOTE (mristin):
            # We check that all properties with suffix ``_id_set`` are popped in
            # ``__getstate__`` and, consequently, set in ``__setstate__``.
            id_set_properties_set = set(cls.id_set_properties)
            popped_attrs_set = set(popped_attrs)
            if not id_set_properties_set.issubset(popped_attrs_set):
                missing_from_popped = id_set_properties_set - popped_attrs_set
                raise AssertionError(
                    f"In class {class_name!r}, the following properties "
                    f"containing runtime ID sets are not being popped in __getstate__: "
                    f"{sorted(missing_from_popped)}."
                )

            if len(cls.getstate.body) < 2:
                raise AssertionError(
                    f"Class {class_name}: __getstate__ should have at least 2 statements"
                )

            # First statement should be state = self.__dict__.copy().
            getstate_first_stmt = cls.getstate.body[0]
            if not (
                isinstance(getstate_first_stmt, ast.Assign)
                and len(getstate_first_stmt.targets) == 1
                and isinstance(getstate_first_stmt.targets[0], ast.Name)
                and getstate_first_stmt.targets[0].id == "state"
            ):
                raise AssertionError(
                    f"In class {class_name!r}, the method __getstate__ should start "
                    f"with 'state = ...'"
                )

            # Last statement should be return state.
            getstate_last_stmt = cls.getstate.body[-1]
            if not (
                isinstance(getstate_last_stmt, ast.Return)
                and isinstance(getstate_last_stmt.value, ast.Name)
                and getstate_last_stmt.value.id == "state"
            ):
                raise AssertionError(
                    f"In class {class_name!r}, the method __getstate__ should end "
                    f"with 'return state'"
                )

            if len(cls.setstate.body) == 0:
                raise AssertionError(
                    f"In class {class_name!r}, the method __setstate__ should have "
                    f"at least 1 statement"
                )

            # First statement should be self.__dict__.update(state).
            setstate_first_stmt = cls.setstate.body[0]
            if not (
                isinstance(setstate_first_stmt, ast.Expr)
                and isinstance(setstate_first_stmt.value, ast.Call)
                and isinstance(setstate_first_stmt.value.func, ast.Attribute)
                and setstate_first_stmt.value.func.attr == "update"
            ):
                raise AssertionError(
                    f"In class {class_name!r}, the method __setstate__ should start "
                    f"with 'self.__dict__.update(state)'"
                )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring

# NOTE (mristin, 2022-10-04):
# This is not really test suite, but a playground to document how streaming XML
# working in Python.
import io
import unittest
import xml.etree.ElementTree as ET


class Test_XML_stream(unittest.TestCase):
    def test_self_closing_element_without_namespace(self) -> None:
        text = """\
<environment/>
"""

        reader = io.StringIO(text)

        writer = io.StringIO()

        for event, element in ET.iterparse(reader, events=["start", "end"]):
            prefix, has_prefix, postfix = element.tag.rpartition("}")

            writer.write(f"event is {event!r}\n")
            writer.write(f"element.tag is {element.tag!r}\n")
            writer.write(
                f"prefix, has_prefix, postfix is {prefix, has_prefix, postfix!r}\n"
            )
            writer.write(f"element.text is {element.text!r}\n")

        self.assertEqual(
            """\
event is 'start'
element.tag is 'environment'
prefix, has_prefix, postfix is ('', '', 'environment')
element.text is None
event is 'end'
element.tag is 'environment'
prefix, has_prefix, postfix is ('', '', 'environment')
element.text is None
""",
            writer.getvalue(),
        )

    def test_cdata_single(self) -> None:
        text = """\
<something>
<![CDATA[crazy < > content]]>
</something>
"""

        reader = io.StringIO(text)

        writer = io.StringIO()

        for event, element in ET.iterparse(reader, events=["start", "end"]):
            writer.write(f"event is {event!r}\n")
            writer.write(f"element.tag is {element.tag!r}\n")
            writer.write(f"element.text is {element.text!r}\n")
            writer.write(f"element.tail is {element.tail!r}\n")

        self.assertEqual(
            """\
event is 'start'
element.tag is 'something'
element.text is '\\ncrazy < > content\\n'
element.tail is None
event is 'end'
element.tag is 'something'
element.text is '\\ncrazy < > content\\n'
element.tail is None
""",
            writer.getvalue(),
        )

    def test_cdata_enclosed(self) -> None:
        text = """\
<something>
<![CDATA[
crazy < > content
]]>
</something>
"""

        reader = io.StringIO(text)

        writer = io.StringIO()

        for event, element in ET.iterparse(reader, events=["start", "end"]):
            writer.write(f"event is {event!r}\n")
            writer.write(f"element.tag is {element.tag!r}\n")
            writer.write(f"element.text is {element.text!r}\n")
            writer.write(f"element.tail is {element.tail!r}\n")

        self.assertEqual(
            """\
event is 'start'
element.tag is 'something'
element.text is '\\n\\ncrazy < > content\\n\\n'
element.tail is None
event is 'end'
element.tag is 'something'
element.text is '\\n\\ncrazy < > content\\n\\n'
element.tail is None
""",
            writer.getvalue(),
        )

    def test_self_closing_element_with_namespace(self) -> None:
        text = """\
<environment xmlns="https://admin-shell.io/aas/3/0/RC02"/>
"""

        reader = io.StringIO(text)

        writer = io.StringIO()

        for event, element in ET.iterparse(reader, events=["start", "end"]):
            prefix, has_prefix, postfix = element.tag.rpartition("}")

            writer.write(f"event is {event!r}\n")
            writer.write(f"element.tag is {element.tag!r}\n")
            writer.write(
                f"prefix, has_prefix, postfix is {prefix, has_prefix, postfix!r}\n"
            )

        self.assertEqual(
            """\
event is 'start'
element.tag is '{https://admin-shell.io/aas/3/0/RC02}environment'
prefix, has_prefix, postfix is ('{https://admin-shell.io/aas/3/0/RC02', '}', 'environment')
event is 'end'
element.tag is '{https://admin-shell.io/aas/3/0/RC02}environment'
prefix, has_prefix, postfix is ('{https://admin-shell.io/aas/3/0/RC02', '}', 'environment')
""",
            writer.getvalue(),
        )

    def test_nested_elements_with_text_and_tail(self) -> None:
        text = """\
<something>some prefix<another>another text</another>some suffix</something>"""

        reader = io.StringIO(text)

        writer = io.StringIO()

        for event, element in ET.iterparse(reader, events=["start", "end"]):
            writer.write(f"event is {event!r}\n")
            writer.write(f"element.tag is {element.tag!r}\n")
            writer.write(f"element.text is {element.text!r}\n")
            writer.write(f"element.tail is {element.tail!r}\n")

        self.assertEqual(
            """\
event is 'start'
element.tag is 'something'
element.text is 'some prefix'
element.tail is None
event is 'start'
element.tag is 'another'
element.text is 'another text'
element.tail is 'some suffix'
event is 'end'
element.tag is 'another'
element.text is 'another text'
element.tail is 'some suffix'
event is 'end'
element.tag is 'something'
element.text is 'some prefix'
element.tail is None
""",
            writer.getvalue(),
        )

    def test_escapes_in_text(self) -> None:
        text = """\
<something>&#x4D;</something>"""

        reader = io.StringIO(text)

        writer = io.StringIO()

        for event, element in ET.iterparse(reader, events=["start", "end"]):
            writer.write(f"event is {event!r}\n")
            writer.write(f"element.tag is {element.tag!r}\n")
            writer.write(f"element.text is {element.text!r}\n")

        self.assertEqual(
            """\
event is 'start'
element.tag is 'something'
element.text is 'M'
event is 'end'
element.tag is 'something'
element.text is 'M'
""",
            writer.getvalue(),
        )

    def test_attrib(self) -> None:
        text = """\
<something that="ok"></something>"""

        reader = io.StringIO(text)

        writer = io.StringIO()

        for event, element in ET.iterparse(reader, events=["start", "end"]):
            writer.write(f"event is {event!r}\n")
            writer.write(f"element.tag is {element.tag!r}\n")
            writer.write(f"element.attrib is {element.attrib!r}\n")

        self.assertEqual(
            """\
event is 'start'
element.tag is 'something'
element.attrib is {'that': 'ok'}
event is 'end'
element.tag is 'something'
element.attrib is {'that': 'ok'}
""",
            writer.getvalue(),
        )

    def test_comment(self) -> None:
        text = """\
<something>
<!-- some comment -->
</something>"""

        reader = io.StringIO(text)

        writer = io.StringIO()

        for event, element in ET.iterparse(reader, events=["start", "end"]):
            writer.write(f"event is {event!r}\n")
            writer.write(f"element.tag is {element.tag!r}\n")
            writer.write(f"element.text is {element.text!r}\n")

        self.assertEqual(
            """\
event is 'start'
element.tag is 'something'
element.text is '\\n\\n'
event is 'end'
element.tag is 'something'
element.text is '\\n\\n'
""",
            writer.getvalue(),
        )


if __name__ == "__main__":
    unittest.main()



class Some_enum(Enum):
    Some_literal = "SOME-LITERAL"
    Another_literal = "ANOTHER-LITERAL"


Some_set: Set[Some_enum] = constant_set(
    values=[Some_enum.Some_literal, Some_enum.Yet_another_literal]
)

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



"""
We encountered a bug when designing V3.0. The length constraints on an inherited
property were not inferred properly.

We add the following test to make sure that the bug does not regress.
"""


@abstract
class Abstract_lang_string(DBC):
    text: str

    def __init__(self, text: str) -> None:
        self.text = text


@invariant(
    lambda self: len(self.text) <= 128,
    "String shall have a maximum length of 128 characters.",
)
class Lang_string_name_type(Abstract_lang_string, DBC):
    def __init__(self, text: str) -> None:
        Abstract_lang_string.__init__(self, text=text)


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



"""
Regression
"""


@invariant(lambda self: len(self) >= 1, "Check if string is not empty.")
@invariant(
    lambda self: matches_XML_serializable_string(self),
    "Constraint AASd-130: An attribute with data type 'string' shall consist "
    "of these characters only: "
    r"^[\x09\x0A\x0D\x20-\uD7FF\uE000-\uFFFD\U00010000-\U0010FFFF]*$",
)
# fmt: on
class Non_empty_XML_serializable_string(str, DBC):
    r"""
    Represent a string with at least one character.

    The string should also be serializable to XML, which is the background for
    the following constraint.

    :constraint AASd-130:

        An attribute with data type "string" shall consist of these characters only:
        ^[\x09\x0A\x0D\x20-\uD7FF\uE000-\uFFFD\u00010000-\u0010FFFF]*$.
    """


@abstract
class Abstract_value(DBC):
    text: Non_empty_XML_serializable_string

    def __init__(self, text: Non_empty_XML_serializable_string) -> None:
        self.text = text


@invariant(
    lambda self: len(self.text) <= 18,
    "String shall have a maximum length of 1023 characters.",
)
class Value(Abstract_value):
    def __init__(self, text: Non_empty_XML_serializable_string) -> None:
        Abstract_value.__init__(self, text=text)


# fmt: off
@invariant(
    lambda self:
    not (self.value is not None)
    or len(self.value) >= 1,
    "Value must be either not set or have at least one item"
)
# fmt: on
class Something:
    value: Optional[Value]

    def __init__(self, value: Optional[Value]) -> None:
        self.value = value


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



class SomethingAbstract:
    property_a: str
    property_b: str

    property_c: Optional[str]
    property_d: Optional[str]

    def __init__(
        self,
        property_a: str,
        property_b: str,
        property_c: Optional[str] = None,
        property_d: Optional[str] = None,
    ) -> None:
        self.property_a = property_a
        self.property_b = property_b
        self.property_c = property_c
        self.property_d = property_d


class Something(SomethingAbstract):
    property_e: str
    property_f: str

    property_g: Optional[str]
    property_h: Optional[str]

    # NOTE (mristin, 2022-03-25):
    # The order of the inherited and defined properties do not match the order of
    # the constructor arguments.

    def __init__(
        self,
        property_e: str,
        property_f: str,
        property_a: str,
        property_b: str,
        property_g: Optional[str] = None,
        property_h: Optional[str] = None,
        property_c: Optional[str] = None,
        property_d: Optional[str] = None,
    ) -> None:
        SomethingAbstract.__init__(
            self,
            property_a=property_a,
            property_b=property_b,
            property_c=property_c,
            property_d=property_d,
        )

        self.property_e = property_e
        self.property_f = property_f
        self.property_g = property_g
        self.property_h = property_h


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



class Something:
    def __init__(self) -> None:
        pass

    @non_mutating
    @implementation_specific
    def get_something(self) -> int:
        return 4


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



"""Check the JSON schema against the given JSON file."""
import argparse
import json
import pathlib
import sys
import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    # NOTE (mristin):
    # We need to disable warnings. Jsonschema package at the latest version (4.4.0) has
    # a problem with JSON schema draft 2019-09 and crashes with a recursion error,
    # see: https://github.com/python-jsonschema/jsonschema/issues/847.
    #
    # We revert to jsonschema 3.2.0, which can not handle 2019-09, but still seems
    # to validate correctly our examples.
    import jsonschema


def main() -> None:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--model_path",
        help="Path to a model to be validated against the generated JSON schema",
        required=True,
    )
    parser.add_argument("--schema_path", help="Path to the JSON schema", required=True)
    args = parser.parse_args()
    model_path = pathlib.Path(args.model_path)
    schema_path = pathlib.Path(args.schema_path)

    if not model_path.exists():
        print(f"--model_path does not exist: {model_path}", file=sys.stderr)
        sys.exit(1)

    if not schema_path.exists():
        print(f"--schema_path does not exist: {schema_path}", file=sys.stderr)
        sys.exit(1)

    with schema_path.open("rt") as fid:
        schema = json.load(fid)

    with model_path.open("rt") as fid:
        instance = json.load(fid)

    try:
        jsonschema.validate(instance=instance, schema=schema)
    except jsonschema.ValidationError as err:
        raise AssertionError(
            f"Failed to validate {model_path} against {schema_path}"
        ) from err


if __name__ == "__main__":
    main()



# pylint: disable=missing-docstring

import textwrap
import unittest

import tests.common
from aas_core_codegen import intermediate
from aas_core_codegen.common import Identifier


class TestMappingOfPrimitiveTypesProperlyExposed(unittest.TestCase):
    def test_primitive_type_to_python_type(self) -> None:
        self.assertEqual(
            intermediate.PRIMITIVE_TYPE_TO_PYTHON_TYPE[intermediate.PrimitiveType.BOOL],
            bool,
        )

        self.assertEqual(
            intermediate.PRIMITIVE_TYPE_TO_PYTHON_TYPE[intermediate.PrimitiveType.INT],
            int,
        )

        self.assertEqual(
            intermediate.PRIMITIVE_TYPE_TO_PYTHON_TYPE[
                intermediate.PrimitiveType.FLOAT
            ],
            float,
        )

        self.assertEqual(
            intermediate.PRIMITIVE_TYPE_TO_PYTHON_TYPE[intermediate.PrimitiveType.STR],
            str,
        )

        self.assertEqual(
            intermediate.PRIMITIVE_TYPE_TO_PYTHON_TYPE[
                intermediate.PrimitiveType.BYTEARRAY
            ],
            bytearray,
        )

    def test_python_type_to_primitive_type(self) -> None:
        self.assertEqual(
            intermediate.PYTHON_TYPE_TO_PRIMITIVE_TYPE[bool],
            intermediate.PrimitiveType.BOOL,
        )

        self.assertEqual(
            intermediate.PYTHON_TYPE_TO_PRIMITIVE_TYPE[int],
            intermediate.PrimitiveType.INT,
        )

        self.assertEqual(
            intermediate.PYTHON_TYPE_TO_PRIMITIVE_TYPE[float],
            intermediate.PrimitiveType.FLOAT,
        )

        self.assertEqual(
            intermediate.PYTHON_TYPE_TO_PRIMITIVE_TYPE[str],
            intermediate.PrimitiveType.STR,
        )

        self.assertEqual(
            intermediate.PYTHON_TYPE_TO_PRIMITIVE_TYPE[bytearray],
            intermediate.PrimitiveType.BYTEARRAY,
        )


class TestIsSubclassOf(unittest.TestCase):
    def test_no_inheritances(self) -> None:
        source = textwrap.dedent(
            """\
            class Concrete:
                pass


            class AnotherConcrete:
                pass


            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)

        assert symbol_table is not None

        concrete = symbol_table.must_find_concrete_class(Identifier("Concrete"))

        another_concrete = symbol_table.must_find_concrete_class(
            Identifier("AnotherConcrete")
        )

        self.assertTrue(concrete.is_subclass_of(cls=concrete))
        self.assertFalse(concrete.is_subclass_of(cls=another_concrete))

    def test_one_level_ancestor(self) -> None:
        source = textwrap.dedent(
            """\
            class Parent:
                pass


            class Concrete(Parent):
                pass


            class AnotherConcrete:
                pass


            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)

        assert symbol_table is not None

        parent = symbol_table.must_find_concrete_class(Identifier("Parent"))

        concrete = symbol_table.must_find_concrete_class(Identifier("Concrete"))

        another_concrete = symbol_table.must_find_concrete_class(
            Identifier("AnotherConcrete")
        )

        self.assertTrue(concrete.is_subclass_of(cls=concrete))
        self.assertTrue(concrete.is_subclass_of(cls=parent))
        self.assertFalse(concrete.is_subclass_of(cls=another_concrete))

    def test_two_level_ancestor(self) -> None:
        source = textwrap.dedent(
            """\
            class GrandParent:
                pass


            class Parent(GrandParent):
                pass


            class Concrete(Parent):
                pass


            class AnotherConcrete:
                pass


            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)

        assert symbol_table is not None

        grand_parent = symbol_table.must_find_concrete_class(Identifier("GrandParent"))

        parent = symbol_table.must_find_concrete_class(Identifier("Parent"))

        concrete = symbol_table.must_find_concrete_class(Identifier("Concrete"))

        another_concrete = symbol_table.must_find_concrete_class(
            Identifier("AnotherConcrete")
        )

        self.assertTrue(concrete.is_subclass_of(cls=concrete))
        self.assertTrue(concrete.is_subclass_of(cls=parent))
        self.assertTrue(concrete.is_subclass_of(cls=grand_parent))
        self.assertFalse(concrete.is_subclass_of(cls=another_concrete))

    def test_common_ancestor_but_no_subclass(self) -> None:
        source = textwrap.dedent(
            """\
            class Parent:
                pass


            class Concrete(Parent):
                pass


            class AnotherConcrete(Parent):
                pass


            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)

        assert symbol_table is not None

        parent = symbol_table.must_find_concrete_class(Identifier("Parent"))

        concrete = symbol_table.must_find_concrete_class(Identifier("Concrete"))

        another_concrete = symbol_table.must_find_concrete_class(
            Identifier("AnotherConcrete")
        )

        self.assertTrue(concrete.is_subclass_of(cls=concrete))
        self.assertTrue(concrete.is_subclass_of(cls=parent))
        self.assertFalse(concrete.is_subclass_of(cls=another_concrete))


class TestMustFindConstant(unittest.TestCase):
    def test_empty(self) -> None:
        source = """\
__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        with self.assertRaises(KeyError):
            _ = symbol_table.must_find_constant(Identifier("Something"))

    def test_constant(self) -> None:
        source = """\
Something: int = constant_int(value=1984)

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        result = symbol_table.must_find_constant(Identifier("Something"))
        assert isinstance(result, intermediate.Constant)

    def test_constant_primitive(self) -> None:
        source = """\
Something: int = constant_int(value=1984)

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        result = symbol_table.must_find_constant_primitive(Identifier("Something"))
        assert isinstance(result, intermediate.ConstantPrimitive)

        with self.assertRaises(TypeError):
            _ = symbol_table.must_find_constant_set_of_primitives(
                Identifier("Something")
            )

        with self.assertRaises(TypeError):
            _ = symbol_table.must_find_constant_set_of_enumeration_literals(
                Identifier("Something")
            )

    def test_constant_set_of_primitives(self) -> None:
        source = """\
Something: Set[str] = constant_set(
    values=["hello", "world"]
)

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
        """
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        result = symbol_table.must_find_constant_set_of_primitives(
            Identifier("Something")
        )
        assert isinstance(result, intermediate.ConstantSetOfPrimitives)

        with self.assertRaises(TypeError):
            _ = symbol_table.must_find_constant_primitive(Identifier("Something"))

        with self.assertRaises(TypeError):
            _ = symbol_table.must_find_constant_set_of_enumeration_literals(
                Identifier("Something")
            )

    def test_constant_set_of_enumeration_literals(self) -> None:
        source = """\
class SomeEnum(Enum):
    Some_literal = "SOME-LITERAL"
    Another_literal = "ANOTHER-LITERAL"
    Yet_another_literal = "YET-ANOTHER-LITERAL"


Something: Set[SomeEnum] = constant_set(
    values=[SomeEnum.Some_literal, SomeEnum.Another_literal]
)

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
        """
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        result = symbol_table.must_find_constant_set_of_enumeration_literals(
            Identifier("Something")
        )
        assert isinstance(result, intermediate.ConstantSetOfEnumerationLiterals)

        with self.assertRaises(TypeError):
            _ = symbol_table.must_find_constant_primitive(Identifier("Something"))

        with self.assertRaises(TypeError):
            _ = symbol_table.must_find_constant_set_of_primitives(
                Identifier("Something")
            )


class TestIsEnumerationLiteralOf(unittest.TestCase):
    def test_enumeration_literal_in_enumeration(self) -> None:
        source = """\
class SomeEnum(Enum):
    Some_literal = "SOME-LITERAL"
    Another_literal = "ANOTHER-LITERAL"

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        some_enum = symbol_table.must_find_enumeration(Identifier("SomeEnum"))
        some_literal = some_enum.literals_by_name["Some_literal"]
        another_literal = some_enum.literals_by_name["Another_literal"]

        self.assertTrue(
            symbol_table.is_enumeration_literal_of(
                literal=some_literal,
                enumeration_or_constant_set_name=Identifier("SomeEnum"),
            )
        )
        self.assertTrue(
            symbol_table.is_enumeration_literal_of(
                literal=another_literal,
                enumeration_or_constant_set_name=Identifier("SomeEnum"),
            )
        )

    def test_enumeration_literal_not_in_different_enumeration(self) -> None:
        source = """\
class SomeEnum(Enum):
    Some_literal = "SOME-LITERAL"

class AnotherEnum(Enum):
    Different_literal = "DIFFERENT-LITERAL"

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        some_enum = symbol_table.must_find_enumeration(Identifier("SomeEnum"))
        another_enum = symbol_table.must_find_enumeration(Identifier("AnotherEnum"))

        some_literal = some_enum.literals_by_name["Some_literal"]
        different_literal = another_enum.literals_by_name["Different_literal"]

        self.assertFalse(
            symbol_table.is_enumeration_literal_of(
                literal=some_literal,
                enumeration_or_constant_set_name=Identifier("AnotherEnum"),
            )
        )
        self.assertFalse(
            symbol_table.is_enumeration_literal_of(
                literal=different_literal,
                enumeration_or_constant_set_name=Identifier("SomeEnum"),
            )
        )

    def test_enumeration_literal_in_constant_set(self) -> None:
        source = """\
class SomeEnum(Enum):
    Some_literal = "SOME-LITERAL"
    Another_literal = "ANOTHER-LITERAL"
    Yet_another_literal = "YET-ANOTHER-LITERAL"

SomeSet: Set[SomeEnum] = constant_set(
    values=[SomeEnum.Some_literal, SomeEnum.Another_literal]
)

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        some_enum = symbol_table.must_find_enumeration(Identifier("SomeEnum"))
        some_literal = some_enum.literals_by_name["Some_literal"]
        another_literal = some_enum.literals_by_name["Another_literal"]
        yet_another_literal = some_enum.literals_by_name["Yet_another_literal"]

        # NOTE (mristin):
        # We check here for membership.
        self.assertTrue(
            symbol_table.is_enumeration_literal_of(
                literal=some_literal,
                enumeration_or_constant_set_name=Identifier("SomeSet"),
            )
        )
        self.assertTrue(
            symbol_table.is_enumeration_literal_of(
                literal=another_literal,
                enumeration_or_constant_set_name=Identifier("SomeSet"),
            )
        )

        # NOTE (mristin):
        # We check here for out-of-membership.
        self.assertFalse(
            symbol_table.is_enumeration_literal_of(
                literal=yet_another_literal,
                enumeration_or_constant_set_name=Identifier("SomeSet"),
            )
        )

    def test_type_error_for_concrete_class(self) -> None:
        source = """\
class SomeEnum(Enum):
    Some_literal = "SOME-LITERAL"

class SomeClass:
    pass

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        some_enum = symbol_table.must_find_enumeration(Identifier("SomeEnum"))
        some_literal = some_enum.literals_by_name["Some_literal"]

        with self.assertRaises(TypeError):
            _ = symbol_table.is_enumeration_literal_of(
                literal=some_literal,
                enumeration_or_constant_set_name=Identifier("SomeClass"),
            )

    def test_type_error_for_constant_primitive(self) -> None:
        source = """\
class SomeEnum(Enum):
    Some_literal = "SOME-LITERAL"

SomeConstant: int = constant_int(value=42)

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        some_enum = symbol_table.must_find_enumeration(Identifier("SomeEnum"))
        some_literal = some_enum.literals_by_name["Some_literal"]

        with self.assertRaises(TypeError):
            _ = symbol_table.is_enumeration_literal_of(
                literal=some_literal,
                enumeration_or_constant_set_name=Identifier("SomeConstant"),
            )

    def test_type_error_for_constant_set_of_primitives(self) -> None:
        source = """\
class SomeEnum(Enum):
    Some_literal = "SOME-LITERAL"

SomeSet: Set[str] = constant_set(
    values=["hello", "world"]
)

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        some_enum = symbol_table.must_find_enumeration(Identifier("SomeEnum"))
        some_literal = some_enum.literals_by_name["Some_literal"]

        # Test that TypeError is raised when name refers to a constant set of primitives
        with self.assertRaises(TypeError):
            _ = symbol_table.is_enumeration_literal_of(
                literal=some_literal,
                enumeration_or_constant_set_name=Identifier("SomeSet"),
            )


class TestLiteralValueSet(unittest.TestCase):
    def test_enumeration(self) -> None:
        source = """\
class SomeEnum(Enum):
    Some_literal = "SOME-LITERAL"
    Another_literal = "ANOTHER-LITERAL"
    Yet_another_literal = "YET-ANOTHER-LITERAL"

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        some_enum = symbol_table.must_find_enumeration(Identifier("SomeEnum"))

        expected_literal_values = frozenset(
            ["SOME-LITERAL", "ANOTHER-LITERAL", "YET-ANOTHER-LITERAL"]
        )

        self.assertEqual(expected_literal_values, some_enum.literal_value_set)

    def test_constant_set_of_primitives(self) -> None:
        source = """\
SomeSet: Set[str] = constant_set(
    values=["hello", "world", "test"]
)

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        some_set = symbol_table.must_find_constant_set_of_primitives(
            Identifier("SomeSet")
        )

        expected_literal_values = frozenset(["hello", "world", "test"])

        self.assertEqual(expected_literal_values, some_set.literal_value_set)

    def test_constant_set_of_enumeration_literals(self) -> None:
        source = """\
class SomeEnum(Enum):
    Some_literal = "SOME-LITERAL"
    Another_literal = "ANOTHER-LITERAL"
    Yet_another_literal = "YET-ANOTHER-LITERAL"

SomeSet: Set[SomeEnum] = constant_set(
    values=[SomeEnum.Some_literal, SomeEnum.Another_literal]
)

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
"""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        some_set = symbol_table.must_find_constant_set_of_enumeration_literals(
            Identifier("SomeSet")
        )

        expected_literal_values = frozenset(["SOME-LITERAL", "ANOTHER-LITERAL"])

        self.assertEqual(expected_literal_values, some_set.literal_value_set)


if __name__ == "__main__":
    unittest.main()



@serialization(with_model_type=True)
class Something:
    some_str: str

    def __init__(self, some_str: str) -> None:
        self.some_str = some_str


class More_concrete(Something):
    another_str: str

    def __init__(self, some_str: str, another_str: str) -> None:
        Something.__init__(self, some_str)
        self.another_str = another_str


class Container:
    something: Something
    list_of_somethings: List[Something]

    def __init__(
        self, something: Something, list_of_somethings: List[Something]
    ) -> None:
        self.something = something
        self.list_of_somethings = list_of_somethings


__version__ = "V198.4"

__xml_namespace__ = "https://dummy/198/4"



# pylint: disable=missing-docstring

import unittest

from aas_core_codegen.parse import retree as parse_retree
from aas_core_codegen.cpp import pattern as cpp_pattern


class TestForRealWorldPatterns(unittest.TestCase):
    def test_for_xs_date_prefix(self) -> None:
        pattern = "^(-?[0-9]+)-(0[1-9]|1[0-2])-(0[0-9]|1[0-9]|2[0-9]|30|31).*$"

        regex, error = parse_retree.parse([pattern])
        assert error is None, f"{error=}"
        assert regex is not None

        code = cpp_pattern._generate._generate_program_definition_for_regex(regex=regex)
        self.assertEqual(
            """\
std::vector<std::unique_ptr<revm::Instruction> > program;

{  // ^(-?[0-9]+)-(0[1-9]|1[0-2])-(0[0-9]|1[0-9]|2[0-9]|30|31).*$
  {  // -?[0-9]+
    {  // -?
      program.emplace_back(
        common::make_unique<revm::InstructionSplit>(1, 2)
      );
      // -
      program.emplace_back(  // 1
        common::make_unique<revm::InstructionChar>(L'-')
      );
    }  // -?
    {  // [0-9]+
      // [0-9]
      program.emplace_back(  // 2
        common::make_unique<revm::InstructionSet>(
          std::vector<revm::Range>{
            revm::Range(L'0', L'9')
          }
        )
      );
      program.emplace_back(
        common::make_unique<revm::InstructionSplit>(2, 4)
      );
    }  // [0-9]+
  }  // -?[0-9]+
  // -
  program.emplace_back(  // 4
    common::make_unique<revm::InstructionChar>(L'-')
  );
  {  // 0[1-9]|1[0-2]
    program.emplace_back(
      common::make_unique<revm::InstructionSplit>(6, 9)
    );
    {  // 0[1-9]
      // 0
      program.emplace_back(  // 6
        common::make_unique<revm::InstructionChar>(L'0')
      );
      // [1-9]
      program.emplace_back(
        common::make_unique<revm::InstructionSet>(
          std::vector<revm::Range>{
            revm::Range(L'1', L'9')
          }
        )
      );
    }  // 0[1-9]
    program.emplace_back(
      common::make_unique<revm::InstructionJump>(11)
    );
    {  // 1[0-2]
      // 1
      program.emplace_back(  // 9
        common::make_unique<revm::InstructionChar>(L'1')
      );
      // [0-2]
      program.emplace_back(
        common::make_unique<revm::InstructionSet>(
          std::vector<revm::Range>{
            revm::Range(L'0', L'2')
          }
        )
      );
    }  // 1[0-2]
  }  // 0[1-9]|1[0-2]
  // -
  program.emplace_back(  // 11
    common::make_unique<revm::InstructionChar>(L'-')
  );
  {  // 0[0-9]|1[0-9]|2[0-9]|30|31
    program.emplace_back(
      common::make_unique<revm::InstructionSplit>(13, 16)
    );
    {  // 0[0-9]
      // 0
      program.emplace_back(  // 13
        common::make_unique<revm::InstructionChar>(L'0')
      );
      // [0-9]
      program.emplace_back(
        common::make_unique<revm::InstructionSet>(
          std::vector<revm::Range>{
            revm::Range(L'0', L'9')
          }
        )
      );
    }  // 0[0-9]
    program.emplace_back(
      common::make_unique<revm::InstructionJump>(30)
    );
    program.emplace_back(  // 16
      common::make_unique<revm::InstructionSplit>(17, 20)
    );
    {  // 1[0-9]
      // 1
      program.emplace_back(  // 17
        common::make_unique<revm::InstructionChar>(L'1')
      );
      // [0-9]
      program.emplace_back(
        common::make_unique<revm::InstructionSet>(
          std::vector<revm::Range>{
            revm::Range(L'0', L'9')
          }
        )
      );
    }  // 1[0-9]
    program.emplace_back(
      common::make_unique<revm::InstructionJump>(30)
    );
    program.emplace_back(  // 20
      common::make_unique<revm::InstructionSplit>(21, 24)
    );
    {  // 2[0-9]
      // 2
      program.emplace_back(  // 21
        common::make_unique<revm::InstructionChar>(L'2')
      );
      // [0-9]
      program.emplace_back(
        common::make_unique<revm::InstructionSet>(
          std::vector<revm::Range>{
            revm::Range(L'0', L'9')
          }
        )
      );
    }  // 2[0-9]
    program.emplace_back(
      common::make_unique<revm::InstructionJump>(30)
    );
    program.emplace_back(  // 24
      common::make_unique<revm::InstructionSplit>(25, 28)
    );
    {  // 30
      // 3
      program.emplace_back(  // 25
        common::make_unique<revm::InstructionChar>(L'3')
      );
      // 0
      program.emplace_back(
        common::make_unique<revm::InstructionChar>(L'0')
      );
    }  // 30
    program.emplace_back(
      common::make_unique<revm::InstructionJump>(30)
    );
    {  // 31
      // 3
      program.emplace_back(  // 28
        common::make_unique<revm::InstructionChar>(L'3')
      );
      // 1
      program.emplace_back(
        common::make_unique<revm::InstructionChar>(L'1')
      );
    }  // 31
  }  // 0[0-9]|1[0-9]|2[0-9]|30|31
  program.emplace_back(  // 30
    common::make_unique<revm::InstructionMatch>()
  );
}  // ^(-?[0-9]+)-(0[1-9]|1[0-2])-(0[0-9]|1[0-9]|2[0-9]|30|31).*$""",
            code,
        )


if __name__ == "__main__":
    unittest.main()



Some_set: Set[str] = constant_set(
    values=[
        "Hello",
        "World",
        "Tonight",
    ]
)

Another_set: Set[str] = constant_set(
    values=[
        "Hello",
        "World",
    ],
    superset_of=[Some_set],
)

__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



# The parser accepts properties and methods in implementation-specific entities
# although the code generator will disregard them.
#
# This is useful, for example, for linting or for visualizing the meta-model
# even though no code is generated based on these properties and methods.


@implementation_specific
class Something:
    x: int

    def do_something(self) -> None:
        pass


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



"""Encapsulate the entry point to different generators."""
import io
import pathlib
import textwrap
from typing import Sequence, TextIO, Tuple, Optional

import asttokens
from icontract import require, ensure

from aas_core_codegen import specific_implementations, intermediate, parse
from aas_core_codegen.common import LinenoColumner


class Context:
    """Represent the context of a code generation."""

    @require(lambda model_path: model_path.exists() and model_path.is_file())
    @require(lambda output_dir: output_dir.exists() and output_dir.is_dir())
    def __init__(
        self,
        model_path: pathlib.Path,
        symbol_table: intermediate.SymbolTable,
        spec_impls: specific_implementations.SpecificImplementations,
        lineno_columner: LinenoColumner,
        output_dir: pathlib.Path,
    ) -> None:
        """Initialize with the given values."""
        self.model_path = model_path
        self.symbol_table = symbol_table
        self.spec_impls = spec_impls
        self.lineno_columner = lineno_columner
        self.output_dir = output_dir


@require(
    lambda errors: all(
        len(error) > 0 and not error.startswith("\n")
        # This is necessary so that we do not have double bullet point.
        and not error.startswith("*") and not error.endswith("\n")
        for error in errors
    )
)
@require(lambda message: not message.endswith(":"))
@require(lambda message: not message.endswith("\n"))
@require(lambda message: not message.startswith("\n") and not message.startswith("*"))
# fmt: on
def write_error_report(message: str, errors: Sequence[str], stderr: TextIO) -> None:
    """
    Write the report (main ``message`` and details as ``errors``) to ``stderr``.

    This method helps us to have a unified way of showing errors.
    """
    stderr.write(f"{message}:\n")
    for error in errors:
        indented = textwrap.indent(error, "  ")
        indented = "* " + indented[2:]
        stderr.write(f"{indented}\n")


@require(lambda model_path: model_path.exists() and model_path.is_file())
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def load_model(
    model_path: pathlib.Path,
) -> Tuple[
    Optional[Tuple[intermediate.SymbolTable, asttokens.ASTTokens]], Optional[str]
]:
    """Load the given meta-model from the file system and understand it."""
    text = model_path.read_text(encoding="utf-8")

    atok, parse_exception = parse.source_to_atok(source=text)
    if parse_exception:
        if isinstance(parse_exception, SyntaxError):
            return None, (
                f"Failed to parse the meta-model: "
                f"invalid syntax at line {parse_exception.lineno}"
            )
        else:
            return None, f"Failed to parse the meta-model: {parse_exception}"

    assert atok is not None

    import_errors = parse.check_expected_imports(atok=atok)
    if import_errors:
        writer = io.StringIO()
        write_error_report(
            message="One or more unexpected imports in the meta-model",
            errors=import_errors,
            stderr=writer,
        )
        return None, writer.getvalue()

    lineno_columner = LinenoColumner(atok=atok)

    parsed_symbol_table, error = parse.atok_to_symbol_table(atok=atok)
    if error is not None:
        writer = io.StringIO()
        write_error_report(
            message="Failed to construct the symbol table",
            errors=[lineno_columner.error_message(error)],
            stderr=writer,
        )
        return None, writer.getvalue()

    assert parsed_symbol_table is not None

    ir_symbol_table, error = intermediate.translate(
        parsed_symbol_table=parsed_symbol_table,
        atok=atok,
    )
    if error is not None:
        writer = io.StringIO()
        write_error_report(
            message=(
                "Failed to translate the parsed symbol table "
                "to intermediate symbol table"
            ),
            errors=[lineno_columner.error_message(error)],
            stderr=writer,
        )
        return None, writer.getvalue()

    assert ir_symbol_table is not None

    return (ir_symbol_table, atok), None



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

import ast
import os
import pathlib
import unittest
from typing import Tuple, List, Union, Sequence

import asttokens

from aas_core_codegen.common import assert_never
from aas_core_codegen.parse import (
    tree as parse_tree,
    _rules as parse_rules,
    retree as parse_retree,
)

import tests.common


def parse_values_from_source(
    root_node: ast.AST,
) -> Sequence[Union[str, parse_tree.FormattedValue]]:
    """Extract the values from the ``root_node`` representing a regular expression."""
    assert isinstance(
        root_node, ast.Module
    ), f"Expected a module, but got: {type(root_node)=}"
    assert (
        len(root_node.body) == 1
    ), f"Expected only a single statement, but got: {ast.dump(root_node)}"

    expr_node = root_node.body[0]
    assert isinstance(expr_node, ast.Expr), (
        f"Expected only an expression in the module body, "
        f"but got: {type(expr_node)=}"
    )

    value_node = expr_node.value

    our_node, rule_error = parse_rules.ast_node_to_our_node(node=value_node)
    assert rule_error is None, (
        f"Unexpected translation error at the parse stage: "
        f"{tests.common.most_underlying_messages(rule_error)}"
    )
    del value_node  # Delete so that we do not use it anymore by mistake

    values = []  # type: Sequence[Union[str, parse_tree.FormattedValue]]

    if isinstance(our_node, parse_tree.Constant):
        assert isinstance(our_node.value, str), (
            f"Expected either a string constant or a joined string, "
            f"but got: {type(our_node.value)=}"
        )
        values = [our_node.value]

    elif isinstance(our_node, parse_tree.JoinedStr):
        values = our_node.values

    else:
        raise AssertionError(
            f"Expected either a string constant or a joined string, "
            f"but got: {type(our_node)=}"
        )

    return values


class Test_cursor(unittest.TestCase):
    def test_empty_values(self) -> None:
        cursor = parse_retree.Cursor(values=[])
        self.assertTrue(cursor.done())

    def test_empty_string(self) -> None:
        source = '""'
        atok = asttokens.ASTTokens(source_text=source, parse=True)
        assert atok.tree is not None
        values = parse_values_from_source(root_node=atok.tree)

        cursor = parse_retree.Cursor(values=values)
        self.assertTrue(cursor.done())

    def test_iteration(self) -> None:
        table = [
            ('"abc"', ["a", "b", "c"]),
            ('f"{x}"', ["<formatted value>"]),
            ('f"{x}ab"', ["<formatted value>", "a", "b"]),
            ('f"a{x}b"', ["a", "<formatted value>", "b"]),
            ('f"ab{x}"', ["a", "b", "<formatted value>"]),
        ]

        for source, expected_tokens in table:
            atok = asttokens.ASTTokens(source_text=source, parse=True)
            assert atok.tree is not None
            values = parse_values_from_source(root_node=atok.tree)

            cursor = parse_retree.Cursor(values=values)

            tokens = []  # type: List[str]
            while not cursor.done():
                formatted_value = cursor.try_formatted_value()
                if formatted_value:
                    tokens.append("<formatted value>")
                else:
                    character = cursor.try_substring(length=1)
                    assert character is not None
                    tokens.append(character)

            self.assertListEqual(expected_tokens, tokens, f"{source=}")

    def test_try_string_in_formatted_value(self) -> None:
        source = 'f"{x}"'
        atok = asttokens.ASTTokens(source_text=source, parse=True)
        assert atok.tree is not None
        values = parse_values_from_source(root_node=atok.tree)

        cursor = parse_retree.Cursor(values=values)

        assert not cursor.try_literal("a")
        self.assertIsNone(cursor.try_substring(length=1))

        assert not cursor.done()

        assert cursor.try_formatted_value() is not None
        assert cursor.done()

    def test_try_jump_over_to_formatted_value(self) -> None:
        source = 'f"abc{x}"'
        atok = asttokens.ASTTokens(source_text=source, parse=True)
        assert atok.tree is not None
        values = parse_values_from_source(root_node=atok.tree)

        cursor = parse_retree.Cursor(values=values)

        assert not cursor.try_literal("abcd")
        self.assertIsNone(cursor.try_substring(length=4))

        assert not cursor.done()


class Test_against_recorded(unittest.TestCase):
    def test_cases(self) -> None:
        repo_root = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent.parent
        test_cases_dir = repo_root / "dev/test_data/parse_retree"

        assert test_cases_dir.exists(), f"{test_cases_dir=}"
        assert test_cases_dir.is_dir(), f"{test_cases_dir=}"

        # The expected cases should have no errors.
        expected_pths = sorted((test_cases_dir / "expected").glob("**/source.py"))
        pths_expected_errors = [
            (pth, False) for pth in expected_pths
        ]  # type: List[Tuple[pathlib.Path, bool]]

        unexpected_pths = sorted((test_cases_dir / "unexpected").glob("**/source.py"))

        pths_expected_errors.extend((pth, True) for pth in unexpected_pths)

        for source_pth, expected_errors in pths_expected_errors:
            case_dir = source_pth.parent

            try:
                source = source_pth.read_text(encoding="utf-8")

                atok = asttokens.ASTTokens(
                    source_text=source, parse=True, filename=str(source_pth)
                )

                assert atok.tree is not None

                values = parse_values_from_source(root_node=atok.tree)

                parsed_regex, error = parse_retree.parse(values=values)

            except Exception as exception:
                raise AssertionError(
                    f"Expected no exception "
                    f"for the test case {case_dir.relative_to(test_cases_dir)}"
                ) from exception

            if not expected_errors and error is not None:
                regex_line, pointer_line = parse_retree.render_pointer(error.cursor)

                raise AssertionError(
                    f"Expected no errors in the test "
                    f"case {case_dir.relative_to(test_cases_dir)}, but got:\n"
                    f"{error.message}\n"
                    f"{regex_line}\n"
                    f"{pointer_line}"
                )

            elif expected_errors and error is None:
                raise AssertionError(
                    f"Expected errors in the test "
                    f"case {case_dir.relative_to(test_cases_dir)}, but got none."
                )

            else:
                pass

            expected_parsed_regex_pth = case_dir / "expected_parsed_regex.txt"
            rendered_regex_pth = case_dir / "rendered_regex.txt"
            expected_error_pth = case_dir / "expected_error.txt"

            if expected_errors:
                if expected_parsed_regex_pth.exists():
                    raise AssertionError(
                        f"Unexpected recorded regex file when errors "
                        f"are expected: {expected_parsed_regex_pth}"
                    )

                assert error is not None

                regex_line, pointer_line = parse_retree.render_pointer(error.cursor)
                error_str = f"{error.message}\n{regex_line}\n{pointer_line}\n"

                if tests.common.RERECORD:
                    expected_error_pth.write_text(error_str, encoding="utf-8")
                else:
                    expected_error_str = expected_error_pth.read_text(encoding="utf-8")
                    self.assertEqual(expected_error_str, error_str, f"{case_dir=}")

            else:
                if expected_error_pth.exists():
                    raise AssertionError(
                        f"Unexpected recorded error file when no errors "
                        f"are expected: {expected_error_pth}"
                    )

                assert parsed_regex is not None

                parsed_regex_str = parse_retree.dump(parsed_regex)

                rendered_parts = []  # type: List[str]
                for value in parse_retree.render(parsed_regex):
                    if isinstance(value, str):
                        rendered_parts.append(value)
                    elif isinstance(value, parse_tree.FormattedValue):
                        rendered_parts.append("<formatted value>")
                    else:
                        assert_never(value)
                rendered_regex_str = "".join(rendered_parts)

                if tests.common.RERECORD:
                    expected_parsed_regex_pth.write_text(
                        parsed_regex_str, encoding="utf-8"
                    )

                    rendered_regex_pth.write_text(rendered_regex_str, encoding="utf-8")
                else:
                    expected_parsed_regex_str = expected_parsed_regex_pth.read_text(
                        encoding="utf-8"
                    )
                    self.assertEqual(
                        expected_parsed_regex_str,
                        parsed_regex_str,
                        f"{case_dir=}",
                    )

                    expected_rendered_regex_str = rendered_regex_pth.read_text(
                        encoding="utf-8"
                    )
                    self.assertEqual(
                        expected_rendered_regex_str,
                        rendered_regex_str,
                        f"{case_dir=}",
                    )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring

import unittest

import aas_core_codegen.typescript.common as typescript_common


class TestStringLiteral(unittest.TestCase):
    def test_empty(self) -> None:
        self.assertEqual('""', typescript_common.string_literal(""))

    def test_no_quotes(self) -> None:
        self.assertEqual('"x"', typescript_common.string_literal("x"))

    def test_double_quotes_but_no_single_quotes(self) -> None:
        self.assertEqual('"a\\"b"', typescript_common.string_literal('a"b'))

    def test_single_quotes_but_no_double_quotes(self) -> None:
        self.assertEqual('"a\'b"', typescript_common.string_literal("a'b"))

    def test_tab(self) -> None:
        self.assertEqual('"a\\tb"', typescript_common.string_literal("a\tb"))

    def test_unset_in_backticks(self) -> None:
        self.assertEqual(
            '"`AAA`"',
            typescript_common.string_literal("`AAA`", in_backticks=False),
        )

        self.assertEqual('"`AAA`"', typescript_common.string_literal("`AAA`"))

    def test_dont_escape_quotes_in_backticks(self) -> None:
        self.assertEqual(
            '"',
            typescript_common.string_literal(
                '"', without_enclosing=True, in_backticks=True
            ),
        )

    def test_escape_backticks_and_dollar_signs_in_backticks(self) -> None:
        self.assertEqual(
            "\\`$\\${something}",
            typescript_common.string_literal(
                "`$${something}", without_enclosing=True, in_backticks=True
            ),
        )


class TestBytesLiteral(unittest.TestCase):
    def test_empty(self) -> None:
        self.assertTupleEqual(
            ("new Uint8Array()", False), typescript_common.bytes_literal(bytearray(b""))
        )

    def test_one(self) -> None:
        self.assertTupleEqual(
            ("new Uint8Array([0x00])", False),
            typescript_common.bytes_literal(bytearray(b"\x00")),
        )

    def test_eight(self) -> None:
        self.assertTupleEqual(
            ("new Uint8Array([0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07])", False),
            typescript_common.bytes_literal(
                bytearray(b"\x00\x01\x02\x03\x04\x05\x06\x07")
            ),
        )

    def test_nine(self) -> None:
        code, multiline = typescript_common.bytes_literal(
            bytearray(b"\x00\x01\x02\x03\x04\x05\x06\x07\x08")
        )

        self.assertEqual(
            """\
new Uint8Array(
  [
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
    0x08
  ]
)""",
            code,
        )
        self.assertTrue(multiline)

    def test_sixteen(self) -> None:
        code, multiline = typescript_common.bytes_literal(
            bytearray(
                b"\x00\x01\x02\x03\x04\x05\x06\x07" b"\x08\x09\x10\x11\x12\x13\x14\x15"
            )
        )

        self.assertEqual(
            """\
new Uint8Array(
  [
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
    0x08, 0x09, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15
  ]
)""",
            code,
        )

        self.assertTrue(multiline)


if __name__ == "__main__":
    unittest.main()



@abstract
class SomethingAbstract:
    some_property: int

    @require(lambda some_property: some_property > 0)
    def __init__(self, some_property: int) -> None:
        self.some_property = some_property

    def some_func(self) -> None:
        pass


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



class SomeConcreteClass:
    """Represent something."""


class SomeContainerClass:
    x: List[SomeConcreteClass]

    def __init__(self, x: List[SomeConcreteClass]) -> None:
        self.x = x


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



@invariant(lambda self: len(self) > 2, "At least 3 characters")
class Something_constrained(str):
    pass


@invariant(lambda self: len(self) < 10, "At most 9 characters.")
class More_constrained(str):
    pass


class Container:
    more_constrained: More_constrained

    def __init__(self, more_constrained: More_constrained) -> None:
        self.more_constrained = more_constrained


__version__ = "V198.4"

__xml_namespace__ = "https://dummy/198/4"



#!/usr/bin/env python3

"""Run pre-commit checks on the repository."""
import argparse
import enum
import os
import pathlib
import shlex
import subprocess
import sys
from typing import Mapping, Optional, Sequence

# pylint: disable=unnecessary-comprehension


class Step(enum.Enum):
    """Enumerate different pre-commit steps."""

    REFORMAT = "reformat"
    ISORT = "isort"
    SSORT = "ssort"
    MYPY = "mypy"
    PYLINT = "pylint"
    TEST = "test"
    DOCTEST = "doctest"
    CHECK_INIT_AND_PYPROJECT_CONSISTENT = "check-init-and-pyproject-consistent"
    CHECK_HELP_IN_README = "check-help-in-readme"
    INTEGRATION_TEST = "integration-test"


def call_and_report(
    verb: str,
    cmd: Sequence[str],
    cwd: Optional[pathlib.Path] = None,
    env: Optional[Mapping[str, str]] = None,
) -> int:
    """
    Wrap a subprocess call with the reporting to STDERR if it failed.

    Return 1 if there is an error and 0 otherwise.
    """
    cmd_str = " ".join(shlex.quote(part) for part in cmd)

    if cwd is not None:
        print(f"Executing from {cwd}: {cmd_str}")
    else:
        print(f"Executing: {cmd_str}")

    exit_code = subprocess.call(cmd, cwd=str(cwd) if cwd is not None else None, env=env)

    if exit_code != 0:
        print(
            f"Failed to {verb} with exit code {exit_code}: {cmd_str}", file=sys.stderr
        )

    return exit_code


def main() -> int:
    """Execute entry_point routine."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--overwrite",
        help="Try to automatically fix the offending files (e.g., by re-formatting).",
        action="store_true",
    )
    parser.add_argument(
        "--select",
        help=(
            "If set, only the selected steps are executed. "
            "This is practical if some of the steps failed and you want to "
            "fix them in isolation. "
            "The steps are given as a space-separated list of: "
            + " ".join(value.value for value in Step)
        ),
        metavar="",
        nargs="+",
        choices=[value.value for value in Step],
    )
    parser.add_argument(
        "--skip",
        help=(
            "If set, skips the specified steps. "
            "This is practical if some of the steps passed and "
            "you want to fix the remainder in isolation. "
            "The steps are given as a space-separated list of: "
            + " ".join(value.value for value in Step)
        ),
        metavar="",
        nargs="+",
        choices=[value.value for value in Step],
    )

    args = parser.parse_args()

    overwrite = bool(args.overwrite)

    selects = (
        [Step(value) for value in args.select]
        if args.select is not None
        else [value for value in Step]
    )
    skips = [Step(value) for value in args.skip] if args.skip is not None else []

    repo_root = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent

    if Step.REFORMAT in selects and Step.REFORMAT not in skips:
        print("Re-formatting...")
        reformat_targets = [
            "aas_core_codegen",
            "dev/continuous_integration",
            "dev/dev_scripts",
            "dev/integration_tests",
            "dev/tests",
            "dev/test_data",
        ]

        # NOTE (mristin, 2022-10-29):
        # We exclude the snippets and the expected output from the formatting as
        # this causes too many unpredictable results and thus makes the generator
        # harder to debug.
        exclude = ["dev/test_data/python", "dev/test_data/python_protobuf"]

        if overwrite:
            exit_code = call_and_report(
                verb="black",
                cmd=[sys.executable, "-m", "black"]
                + reformat_targets
                + ["--exclude"]
                + exclude,
                cwd=repo_root,
            )
            if exit_code != 0:
                return 1
        else:
            exit_code = call_and_report(
                verb="check with black",
                cmd=[sys.executable, "-m", "black", "--check"]
                + reformat_targets
                + ["--exclude"]
                + exclude,
                cwd=repo_root,
            )
            if exit_code != 0:
                return 1
    else:
        print("Skipped re-formatting.")

    if Step.SSORT in selects and Step.SSORT not in skips:
        print("Ssort'ing...")

        ssort_targets = [
            "aas_core_codegen",
            "dev/continuous_integration",
            "dev/dev_scripts",
            "dev/tests",
        ]

        if overwrite:
            exit_code = call_and_report(
                verb="ssort",
                cmd=[sys.executable, "-m", "ssort"] + ssort_targets,
                cwd=repo_root,
            )
            if exit_code != 0:
                return 1

            # NOTE (mristin, 2024-01-26):
            # We have to re-format just after the ssort,
            # see: https://pypi.org/project/ssort/,
            # "We recommend that you reformat using isort and black
            # immediately after running ssort."
            exit_code = call_and_report(
                verb="black after ssort",
                cmd=[sys.executable, "-m", "black"] + ssort_targets,
                cwd=repo_root,
            )
            if exit_code != 0:
                return 1
        else:
            exit_code = call_and_report(
                verb="check with ssort",
                cmd=[sys.executable, "-m", "ssort", "--check", "--diff"]
                + ssort_targets,
                cwd=repo_root,
            )
            if exit_code != 0:
                return 1

    if Step.MYPY in selects and Step.MYPY not in skips:
        print("Mypy'ing...")
        mypy_targets = [
            "aas_core_codegen",
            "dev/tests",
            "dev/continuous_integration",
            "dev/dev_scripts",
        ]
        config_file = pathlib.Path("dev/continuous_integration") / "mypy.ini"

        exit_code = call_and_report(
            verb="mypy",
            cmd=[
                sys.executable,
                "-m",
                "mypy",
                "--strict",
                "--config-file",
                str(config_file),
            ]
            + mypy_targets,
            cwd=repo_root,
        )
        if exit_code != 0:
            return 1
    else:
        print("Skipped mypy'ing.")

    if Step.PYLINT in selects and Step.PYLINT not in skips:
        print("Pylint'ing...")
        pylint_targets = [
            "aas_core_codegen",
            "dev/tests",
            "dev/continuous_integration",
            "dev/dev_scripts",
        ]
        rcfile = pathlib.Path("dev/continuous_integration") / "pylint.rc"

        exit_code = call_and_report(
            verb="pylint",
            cmd=[sys.executable, "-m", "pylint", f"--rcfile={rcfile}"] + pylint_targets,
            cwd=repo_root,
        )
        if exit_code != 0:
            return 1
    else:
        print("Skipped pylint'ing.")

    if Step.TEST in selects and Step.TEST not in skips:
        print("Testing...")
        env = os.environ.copy()
        env["ICONTRACT_SLOW"] = "true"

        exit_code = call_and_report(
            verb="execute unit tests",
            cmd=[
                sys.executable,
                "-m",
                "coverage",
                "run",
                "--source",
                "aas_core_codegen",
                "-m",
                "unittest",
                "discover",
                "-s",
                "dev/tests",
            ],
            cwd=repo_root,
            env=env,
        )
        if exit_code != 0:
            return 1

        exit_code = call_and_report(
            verb="report the coverage", cmd=["coverage", "report"]
        )
        if exit_code != 0:
            return 1
    else:
        print("Skipped testing.")

    if Step.INTEGRATION_TEST in selects and Step.INTEGRATION_TEST not in skips:
        print("Integration testing...")
        exit_code = call_and_report(
            verb="execute unit tests",
            cmd=[
                sys.executable,
                "dev/integration_tests/main.py",
            ],
            cwd=repo_root,
        )
        if exit_code != 0:
            return 1
    else:
        print("Skipped integration testing.")

    if Step.DOCTEST in selects and Step.DOCTEST not in skips:
        print("Doctest'ing...")

        # BEFORE-RELEASE (mristin, 2021-12-13):
        #  Add ``{repo_root}/docs/source/**/*.rst`` as well here
        doc_files = ["README.rst"]

        exit_code = call_and_report(
            verb="doctest",
            cmd=[sys.executable, "-m", "doctest"] + doc_files,
            cwd=repo_root,
        )
        if exit_code != 0:
            return 1

        for pth in (repo_root / "aas_core_codegen").glob("**/*.py"):
            if pth.name == "__main__.py":
                continue

            # NOTE (mristin, 2021-12-27):
            # The subprocess calls are expensive, call only if there is an actual
            # doctest
            text = pth.read_text(encoding="utf-8")
            if ">>>" in text:
                exit_code = call_and_report(
                    verb="doctest",
                    cmd=[sys.executable, "-m", "doctest", str(pth)],
                    cwd=repo_root,
                )
                if exit_code != 0:
                    return 1
    else:
        print("Skipped doctest'ing.")

    if (
        Step.CHECK_INIT_AND_PYPROJECT_CONSISTENT in selects
        and Step.CHECK_INIT_AND_PYPROJECT_CONSISTENT not in skips
    ):
        print(
            "Checking that aas_core_codegen/__init__.py and pyproject.toml "
            "are consistent..."
        )
        exit_code = call_and_report(
            verb=(
                "check that aas_core_codegen/__init__.py and pyproject.toml "
                "are consistent"
            ),
            cmd=[
                sys.executable,
                "dev/continuous_integration/check_init_and_pyproject_consistent.py",
            ],
            cwd=repo_root,
        )
        if exit_code != 0:
            return 1
    else:
        print(
            "Skipped checking that aas_core_codegen/__init__.py and "
            "pyproject.toml are consistent."
        )

    # NOTE (mristin):
    # We need to check for the Python version since ``argparse`` output changes
    # between the versions. Hence, we pin it at the moment to Python 3.12.

    if (3, 12) <= sys.version_info < (3, 13):
        if (
            Step.CHECK_HELP_IN_README in selects
            and Step.CHECK_HELP_IN_README not in skips
        ):
            cmd = [sys.executable, "dev/continuous_integration/check_help_in_readme.py"]
            if overwrite:
                cmd.append("--overwrite")

            if not overwrite:
                print("Checking that --help's and the readme coincide...")
                exit_code = call_and_report(
                    verb="check that --help's and the readme coincide",
                    cmd=[
                        sys.executable,
                        "dev/continuous_integration/check_help_in_readme.py",
                    ],
                    cwd=repo_root,
                )
                if exit_code != 0:
                    return 1
            else:
                print("Overwriting the --help's in the readme...")
                exit_code = call_and_report(
                    verb="overwrite the --help's in the readme",
                    cmd=[
                        sys.executable,
                        "dev/continuous_integration/check_help_in_readme.py",
                        "--overwrite",
                    ],
                    cwd=repo_root,
                )
                if exit_code != 0:
                    return 1

            subprocess.check_call(cmd, cwd=str(repo_root))
        else:
            print("Skipped checking that --help's and the doc coincide.")
    else:
        print(
            "Skipped checking that --help's and the doc coincide "
            "since we pin it on Python version 3.12."
        )

    return 0


if __name__ == "__main__":
    sys.exit(main())



class Something:
    @snapshot(capture=lambda lst: lst.copy(), name="lst")
    @ensure(lambda lst, OLD: lst == OLD.lst)
    def do_something(self, lst: List[int]) -> None:
        pass


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



@verification
def match_something(text: str) -> bool:
    pattern = f"^(\\U00010000*something)?$"
    return match(pattern, text) is not None


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



"""Test end-to-end: generate the SDK and then apply it on concrete models."""

import argparse
import io
import os
import pathlib
import shlex
import shutil
import subprocess
import sys
from typing import Final, Sequence, Iterator, Tuple

import aas_core_codegen.main


class IntegrationTestCase:
    """
    Provide a protocol for running different stages of an integration test.

    It is expected that the code of the SDK has been previously generated.

    There are two stages of an integration test:
    * ``build`` — where the project files, including test drivers, are built, and
    * ``execute`` — where the test driver is executed.
    """

    name: Final[str]
    meta_model_path: Final[pathlib.Path]
    input_dir: Final[pathlib.Path]
    output_dir: Final[pathlib.Path]
    test_data_dir: Final[pathlib.Path]

    def __init__(
        self,
        name: str,
        meta_model_path: pathlib.Path,
        input_dir: pathlib.Path,
        output_dir: pathlib.Path,
        test_data_dir: pathlib.Path,
    ) -> None:
        self.name = name
        self.meta_model_path = meta_model_path
        self.input_dir = input_dir
        self.output_dir = output_dir
        self.test_data_dir = test_data_dir

    def _info(self, message: str) -> None:
        """Log the ``message`` to STDOUT."""
        print(f"[{self.name}] {message}")

    def _panic(self, message: str) -> None:
        """Log the ``message`` to STDERR and exit the program."""
        print(f"[{self.name}] {message}", file=sys.stderr)
        sys.exit(1)

    def _generate(
        self,
        meta_model_path: pathlib.Path,
        target: aas_core_codegen.main.Target,
        target_dir: pathlib.Path,
    ) -> None:
        """
        Generate the SDK.

        This method is expected to be called from within the :py:meth:`setup`.
        """
        params = aas_core_codegen.main.Parameters(
            model_path=meta_model_path,
            target=target,
            snippets_dir=self.input_dir / "snippets",
            # NOTE (mristin):
            # The output directory for aas-core-codegen is where we put the SDK code.
            # The SDK code might not necessarily live at the root of the test driver
            # directory, so we call it here ``target_dir``.
            output_dir=target_dir,
        )

        stdout = io.StringIO()
        stderr = io.StringIO()

        return_code = aas_core_codegen.main.execute(
            params=params, stdout=stdout, stderr=stderr
        )
        if stderr.getvalue() != "":
            self._panic(
                f"Expected no stderr when generating the SDK, "
                f"but got:\n{stderr.getvalue()}"
            )
        if return_code != 0:
            self._panic(
                f"Expected 0 return code when generating the SDK, "
                f"but got: {return_code}"
            )

    def setup(self) -> None:
        """Generate the SDK, copy the boilerplate code and set up the test project."""
        return

    def build(self) -> None:
        """
        Build the test project to be executed in the integration test.

        The default is to build nothing — override this method to specify how to build
        the test project.
        """
        return

    def _must_call(self, command: Sequence[str], cwd: pathlib.Path) -> None:
        """Log the command, execute it and panic if it fails."""
        assert all(
            isinstance(part, str) for part in command
        ), f"Expected all parts of the command to be string, but got command: {command}"

        cmd_joined = " ".join(shlex.quote(part) for part in command)
        self._info(f"Executing: {cmd_joined}")
        return_code = subprocess.call(command, cwd=str(cwd))

        if return_code != 0:
            self._panic(
                f"The command {cmd_joined} failed with return code {return_code}"
            )

    def execute(self) -> None:
        """
        Execute the test driver.

        The default is to execute nothing — override this method to call
        your specific test driver.
        """
        return


def _escaped_and_joined_command(command: Sequence[str]) -> str:
    """Generate the string such that it can be copy/pasted into a terminal."""
    return " ".join(shlex.quote(part) for part in command)


class JsonSchemaIntegrationTest(IntegrationTestCase):
    """Test the generation of JSON schema and validate it against the test data."""

    def setup(self) -> None:
        self._generate(
            meta_model_path=self.meta_model_path,
            target=aas_core_codegen.main.Target.JSONSCHEMA,
            target_dir=self.output_dir,
        )

        shutil.copy(
            str(self.input_dir / "boilerplate" / "main.py"), self.output_dir / "main.py"
        )

    def execute(self) -> None:
        self._must_call(
            command=[
                sys.executable,
                str(self.output_dir / "main.py"),
                "--schema_path",
                str(self.output_dir / "schema.json"),
                "--model_path",
                str(self.test_data_dir / "json" / "full.json"),
            ],
            cwd=self.output_dir,
        )


def _over_model_paths(test_data_dir: pathlib.Path) -> Iterator[pathlib.Path]:
    """Iterate over all the possible model files in the given test data directory."""
    for path in test_data_dir.glob("**/*"):
        if path.suffix.lower() in (".xml", ".json"):
            yield path


class PythonIntegrationTest(IntegrationTestCase):
    def setup(self) -> None:
        self._generate(
            meta_model_path=self.meta_model_path,
            target=aas_core_codegen.main.Target.PYTHON,
            target_dir=self.output_dir / "sdk",
        )

        shutil.copy(
            str(self.input_dir / "boilerplate" / "main.py"), self.output_dir / "main.py"
        )

    def execute(self) -> None:
        for model_path in _over_model_paths(self.test_data_dir):
            self._must_call(
                command=[
                    sys.executable,
                    str(self.output_dir / "main.py"),
                    "--model_path",
                    str(model_path),
                ],
                cwd=self.output_dir,
            )


def main() -> None:
    """Execute the main routine."""
    repo_dir = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent
    integration_tests_dir = repo_dir / "dev/integration_tests"

    meta_model_path = integration_tests_dir / "input" / "meta_model.py"
    test_data_dir = integration_tests_dir / "test_data"

    test_names_to_tests = {
        "jsonschema": JsonSchemaIntegrationTest(
            name="jsonschema",
            meta_model_path=meta_model_path,
            input_dir=integration_tests_dir / "input" / "jsonschema",
            output_dir=integration_tests_dir / "output" / "jsonschema",
            test_data_dir=test_data_dir,
        ),
        "python": PythonIntegrationTest(
            name="python",
            meta_model_path=meta_model_path,
            input_dir=integration_tests_dir / "input" / "python",
            output_dir=integration_tests_dir / "output" / "python",
            test_data_dir=test_data_dir,
        ),
    }

    # NOTE (mristin):
    # We enforce this convention in order to avoid the confusion on the side of
    # the user.
    assert all(
        test_name == test.name for test_name, test in test_names_to_tests.items()
    )

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--test",
        choices=sorted(test_names_to_tests.keys()),
        default=None,
        help="Run only a specific integration test",
    )
    args = parser.parse_args()

    selected_test_names_tests: Sequence[Tuple[str, IntegrationTestCase]]

    if args.test is not None:
        selected_test_names_tests = [(args.test, test_names_to_tests[args.test])]
    else:
        selected_test_names_tests = [
            (test_name, test_names_to_tests[test_name])
            for test_name in sorted(test_names_to_tests.keys())
        ]

    for test_name, test in selected_test_names_tests:
        print(f"Running the test case: {test_name} ...")

        test.setup()
        test.build()
        test.execute()


if __name__ == "__main__":
    main()



# NOTE (mristin, 2021-07-07):
# We explicitly test that the inheritance from an implementation-specific entity
# is possible. We initially rejected the idea, and wanted to disallow such
# an inheritance chain, but changed our mind eventually due to intuition (but no
# hard reason!).


@abstract
@implementation_specific
class Parent:
    pass


@implementation_specific
class Something(Parent):
    pass


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



"""
Draw a bipartite graph on matching lines in two files.

This is helpful when you want to minimize the diffs. Unfortunately, ``diff`` is not
really a good tool when it comes to re-shuffling of the blocks.
"""

import argparse
import pathlib
import sys


def main() -> int:
    """Execute the main routine."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--ours", help="path to the first file", required=True)
    parser.add_argument("--theirs", help="path to the second file", required=True)
    args = parser.parse_args()

    ours_pth = pathlib.Path(args.ours)
    theirs_pth = pathlib.Path(args.theirs)

    our_lines = ours_pth.read_text(encoding="utf-8").splitlines()
    their_lines = theirs_pth.read_text(encoding="utf-8").splitlines()

    their_lines_by_index = {line: i for i, line in enumerate(their_lines)}

    for i, line in enumerate(our_lines):
        their_i = their_lines_by_index.get(line, None)
        if their_i is not None:
            print(f"Ours {i+1:4d} to theirs {their_i:4d}")

    return 0


if __name__ == "__main__":
    sys.exit(main())



# pylint: disable=missing-docstring

import contextlib
import io
import os
import pathlib
import tempfile
import unittest

import aas_core_codegen.main

import tests.common

_TYPES = [
    "AdministrativeInformation",
    "AnnotatedRelationshipElement",
    "AssetAdministrationShell",
    "AssetInformation",
    "BasicEventElement",
    "Blob",
    "Capability",
    "ConceptDescription",
    "DataSpecificationIec61360",
    "EmbeddedDataSpecification",
    "Entity",
    "Environment",
    "EventPayload",
    "Extension",
    "File",
    "Key",
    "LangStringDefinitionTypeIec61360",
    "LangStringNameType",
    "LangStringPreferredNameTypeIec61360",
    "LangStringShortNameTypeIec61360",
    "LangStringTextType",
    "LevelType",
    "MultiLanguageProperty",
    "Operation",
    "OperationVariable",
    "Property",
    "Qualifier",
    "Range",
    "ReferenceElement",
    "Reference",
    "RelationshipElement",
    "Resource",
    "SpecificAssetId",
    "SubmodelElementCollection",
    "SubmodelElementList",
    "Submodel",
    "ValueList",
    "ValueReferencePair",
]

_CONSTANTS = [pathlib.Path("constants/Constants.java")]
_COPYING = [pathlib.Path("copying/Copying.java")]
_ENHANCING = [pathlib.Path(f"enhancing/Enhanced{t}.java") for t in _TYPES] + [
    pathlib.Path("enhancing/Enhanced.java"),
    pathlib.Path("enhancing/Enhancer.java"),
    pathlib.Path("enhancing/Unwrapper.java"),
    pathlib.Path("enhancing/Wrapper.java"),
]
# NOTE (empwilli 2024-03-26): we create generators only for types that have
# non-mandatory attributes.
_GENERATION = [
    pathlib.Path(f"generation/{t}Builder.java")
    for t in _TYPES
    if t
    not in (
        "EmbeddedDataSpecification",
        "Key",
        "LangStringDefinitionTypeIec61360",
        "LangStringNameType",
        "LangStringPreferredNameTypeIec61360",
        "LangStringShortNameTypeIec61360",
        "LangStringTextType",
        "LevelType",
        "OperationVariable",
        "ValueList",
        "ValueReferencePair",
    )
]
_JSONIZATION = [pathlib.Path("jsonization/Jsonization.java")]
_REPORTING = [pathlib.Path("reporting/Reporting.java")]
_STRINGIFICATION = [pathlib.Path("stringification/Stringification.java")]
_VERIFICATION = [pathlib.Path("verification/Verification.java")]
_STRUCTURE = (
    [
        pathlib.Path("types/enums/AasSubmodelElements.java"),
        pathlib.Path("types/enums/AssetKind.java"),
        pathlib.Path("types/enums/DataTypeDefXsd.java"),
        pathlib.Path("types/enums/DataTypeIec61360.java"),
        pathlib.Path("types/enums/Direction.java"),
        pathlib.Path("types/enums/EntityType.java"),
        pathlib.Path("types/enums/KeyTypes.java"),
        pathlib.Path("types/enums/ModellingKind.java"),
        pathlib.Path("types/enums/QualifierKind.java"),
        pathlib.Path("types/enums/ReferenceTypes.java"),
        pathlib.Path("types/enums/StateOfEvent.java"),
    ]
    + [pathlib.Path(f"types/impl/{t}.java") for t in _TYPES]
    + [pathlib.Path(f"types/model/I{t}.java") for t in _TYPES]
)
_VISITATION = [
    pathlib.Path("visitation/AbstractTransformer.java"),
    pathlib.Path("visitation/AbstractTransformerWithContext.java"),
    pathlib.Path("visitation/AbstractVisitor.java"),
    pathlib.Path("visitation/AbstractVisitorWithContext.java"),
    pathlib.Path("visitation/ITransformer.java"),
    pathlib.Path("visitation/ITransformerWithContext.java"),
    pathlib.Path("visitation/IVisitor.java"),
    pathlib.Path("visitation/IVisitorWithContext.java"),
    pathlib.Path("visitation/VisitorThrough.java"),
]
XMLIZATION = [pathlib.Path("xmlization/Xmlization.java")]

GENERATED_FILES = (
    _CONSTANTS
    + _COPYING
    + _ENHANCING
    + _GENERATION
    + _JSONIZATION
    + _REPORTING
    + _STRINGIFICATION
    + _VERIFICATION
    + _STRUCTURE
    + _VISITATION
)


class Test_against_recorded(unittest.TestCase):
    def test_cases(self) -> None:
        repo_dir = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent.parent

        parent_case_dir = repo_dir / "dev" / "test_data" / "java" / "test_main"
        assert parent_case_dir.exists() and parent_case_dir.is_dir(), parent_case_dir

        for model_pth in tests.common.REAL_META_MODEL_PATHS:
            case_dir = parent_case_dir / model_pth.stem
            assert case_dir.is_dir(), case_dir

            snippets_dir = case_dir / "input/snippets"
            assert snippets_dir.exists() and snippets_dir.is_dir(), snippets_dir

            expected_output_dir = case_dir / "expected_output"

            with contextlib.ExitStack() as exit_stack:
                if tests.common.RERECORD:
                    output_dir = expected_output_dir
                    expected_output_dir.mkdir(exist_ok=True, parents=True)
                else:
                    assert (
                        expected_output_dir.exists() and expected_output_dir.is_dir()
                    ), expected_output_dir

                    # pylint: disable=consider-using-with
                    tmp_dir = tempfile.TemporaryDirectory()
                    exit_stack.push(tmp_dir)
                    output_dir = pathlib.Path(tmp_dir.name)

                params = aas_core_codegen.main.Parameters(
                    model_path=model_pth,
                    target=aas_core_codegen.main.Target.JAVA,
                    snippets_dir=snippets_dir,
                    output_dir=output_dir,
                )

                stdout = io.StringIO()
                stderr = io.StringIO()

                return_code = aas_core_codegen.main.execute(
                    params=params, stdout=stdout, stderr=stderr
                )

                if stderr.getvalue() != "":
                    raise AssertionError(
                        f"Expected no stderr on valid models, but got:\n"
                        f"{stderr.getvalue()}"
                    )

                self.assertEqual(
                    0, return_code, "Expected 0 return code on valid models"
                )

                stdout_pth = expected_output_dir / "stdout.txt"
                normalized_stdout = stdout.getvalue().replace(
                    str(output_dir), "<output dir>"
                )

                if tests.common.RERECORD:
                    stdout_pth.write_text(normalized_stdout, encoding="utf-8")
                else:
                    self.assertEqual(
                        normalized_stdout,
                        stdout_pth.read_text(encoding="utf-8"),
                        stdout_pth,
                    )

                for relevant_rel_pth in GENERATED_FILES:
                    expected_pth = expected_output_dir / relevant_rel_pth
                    output_pth = output_dir / relevant_rel_pth

                    if not output_pth.exists():
                        raise FileNotFoundError(
                            f"The output file is missing: {output_pth}"
                        )

                    try:
                        output = output_pth.read_text(encoding="utf-8")
                    except Exception as exception:
                        raise RuntimeError(
                            f"Failed to read the output from {output_pth}"
                        ) from exception

                    if tests.common.RERECORD:
                        expected_pth.write_text(output, encoding="utf-8")
                    else:
                        try:
                            expected_output = expected_pth.read_text(encoding="utf-8")
                        except Exception as exception:
                            raise RuntimeError(
                                f"Failed to read the expected output "
                                f"from {expected_pth}"
                            ) from exception

                        self.assertEqual(
                            expected_output,
                            output,
                            f"The files {expected_pth} and {output_pth} do not match.",
                        )


if __name__ == "__main__":
    unittest.main()



"""Provide code generation for unrolling recursive calls and iterations."""
import abc
import io
import textwrap
from typing import Sequence, List

from icontract import DBC, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import Identifier, assert_never
from aas_core_codegen.csharp.common import INDENT as I


class Node:
    """Represent a node in the unrolling tree."""

    def __init__(self, text: str, children: Sequence["Node"]):
        self.text = text
        self.children = children


def render(node: Node) -> str:
    """
    Render the node recursively.

    >>> render(Node(text='something', children=[]))
    'something'

    >>> render(Node(text='parent', children=[Node(text='child', children=[])]))
    'parent\\n{\\n    child\\n}'

    >>> render(Node(text='parent', children=[
    ...     Node(text='child1', children=[]),
    ...     Node(text='child2', children=[]),
    ... ]))
    'parent\\n{\\n    child1\\n\\n    child2\\n}'
    """
    if len(node.children) == 0:
        return node.text

    writer = io.StringIO()
    writer.write(node.text)
    writer.write("\n{")

    for i, child in enumerate(node.children):
        if i == 0:
            writer.write("\n")
        else:
            writer.write("\n\n")

        writer.write(textwrap.indent(render(child), I))

    writer.write("\n}")

    return writer.getvalue()


class AbstractUnroller(DBC):
    """Generate code to unroll recursion into generic types."""

    # NOTE (mristin, 2022-09-15):
    # We originally wrote this code to unroll both lists *and* dictionaries. Eventually,
    # no dictionaries were included in the meta-model. Due to the lack of time, we could
    # not clean up the implementation of this class. Additionally, it is well possible
    # that the dictionaries will be introduced eventually in the meta-model, so we leave
    # the code as-is at the moment.
    #
    # If you read this comment, have to change this class and the dictionaries are not
    # part of the meta-model yet, please go ahead and clean it up first.

    @staticmethod
    @require(lambda level: level >= 0)
    @require(lambda suffix: suffix in ("Item", "KeyValue"))
    def _loop_var_name(level: int, suffix: str) -> Identifier:
        """
        Generate the name of the loop variable.

        :param level:
            recursion level

            The level 0 implies the first inner loop.
        :param suffix:
            suffix of the loop variable; we distinguish between items in a list and
            key-value-pairs in a dictionary
        :return: generated loop variable name
        """
        if level == 0:
            if suffix == "Item":
                return Identifier(f"an{suffix}")
            else:
                assert suffix == "KeyValue"
                return Identifier(f"a{suffix}")

        elif level == 1:
            return Identifier(f"another{suffix}")
        else:
            return Identifier("yet" + "Yet" * (level - 1) + f"another{suffix}")

    @abc.abstractmethod
    def _unroll_primitive_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.PrimitiveTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[Node]:
        """Generate code for the given specific ``type_annotation``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def _unroll_our_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OurTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[Node]:
        """Generate code for the given specific ``type_annotation``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def _unroll_list_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.ListTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[Node]:
        """Generate code for the given specific ``type_annotation``."""
        raise NotImplementedError()

    @abc.abstractmethod
    def _unroll_optional_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OptionalTypeAnnotation,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[Node]:
        """Generate code for the given specific ``type_annotation``."""
        raise NotImplementedError()

    @require(lambda item_level: item_level >= 0)
    @require(lambda key_value_level: key_value_level >= 0)
    def unroll(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.TypeAnnotationUnion,
        path: List[str],
        item_level: int,
        key_value_level: int,
    ) -> List[Node]:
        """
        Dispatch the given type annotation to unrolling.

        :param unrollee_expr: Expression of the element to be unrolled
        :param type_annotation: Type annotation corresponding to the ``unrollee_expr``
        :param path:
            Path, as code snippets to be joined by "/"to the ``unrollee_expr``
        :param item_level:
            Depth level of the list loops.

            Level 0 indicates the outer loop.
        :param key_value_level:
            Depth level of the key-value-pairs loops.

            Level 0 indicates the outer loop.
        """
        if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
            return self._unroll_primitive_type_annotation(
                unrollee_expr=unrollee_expr,
                type_annotation=type_annotation,
                path=path,
                item_level=item_level,
                key_value_level=key_value_level,
            )

        elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
            return self._unroll_our_type_annotation(
                unrollee_expr=unrollee_expr,
                type_annotation=type_annotation,
                path=path,
                item_level=item_level,
                key_value_level=key_value_level,
            )

        elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
            return self._unroll_list_type_annotation(
                unrollee_expr=unrollee_expr,
                type_annotation=type_annotation,
                path=path,
                item_level=item_level,
                key_value_level=key_value_level,
            )

        elif isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
            return self._unroll_optional_type_annotation(
                unrollee_expr=unrollee_expr,
                type_annotation=type_annotation,
                path=path,
                item_level=item_level,
                key_value_level=key_value_level,
            )
        else:
            assert_never(type_annotation)

        raise AssertionError("Should not have gotten here")



# pylint: disable=missing-docstring

import unittest
from typing import Optional, List

import aas_core_codegen.java.description as java_description


class TestRemoveRedundantP(unittest.TestCase):
    def test_empty(self) -> None:
        tokens = []  # type: List[java_description._Token]

        result = java_description._remove_redundant_p(tokens)
        self.assertListEqual([], result)

    def test_prefix_chain(self) -> None:
        tokens = [
            java_description._TokenP(),
            java_description._TokenP(),
            java_description._TokenText("hello world"),
        ]  # type: List[java_description._Token]

        result = java_description._remove_redundant_p(tokens)
        self.assertListEqual(
            [
                java_description._TokenP(),
                java_description._TokenText("hello world"),
            ],
            result,
        )

    def test_suffix_chain(self) -> None:
        tokens = [
            java_description._TokenP(),
            java_description._TokenText("hello world"),
            java_description._TokenP(),
            java_description._TokenP(),
        ]  # type: List[java_description._Token]

        result = java_description._remove_redundant_p(tokens)
        self.assertListEqual(
            [
                java_description._TokenP(),
                java_description._TokenText("hello world"),
            ],
            result,
        )

    def test_chain_in_the_middle(self) -> None:
        tokens = [
            java_description._TokenP(),
            java_description._TokenText("hello"),
            java_description._TokenP(),
            java_description._TokenP(),
            java_description._TokenText("world"),
        ]  # type: List[java_description._Token]

        result = java_description._remove_redundant_p(tokens)
        self.assertListEqual(
            [
                java_description._TokenP(),
                java_description._TokenText("hello"),
                java_description._TokenP(),
                java_description._TokenText("world"),
            ],
            result,
        )


class TestIndentionMachine(unittest.TestCase):
    def test_empty(self) -> None:
        machine = java_description._IndentionMachine()
        result = machine.render()

        self.assertEqual("", result)

    def test_only_text(self) -> None:
        machine = java_description._IndentionMachine()
        machine.write("hello world")
        result = machine.render()

        self.assertEqual("hello world", result)

    def test_indent_prefix(self) -> None:
        machine = java_description._IndentionMachine()
        machine.indent()
        machine.write("hello world")
        result = machine.render()

        self.assertEqual("  hello world", result)

    def test_indent_dedent(self) -> None:
        machine = java_description._IndentionMachine()

        machine.write("hello\n")
        machine.indent()
        machine.write("wor\nld\n")
        machine.dedent()
        machine.write("!")
        result = machine.render()

        self.assertEqual(
            """\
hello
  wor
  ld
!""",
            result,
        )

    def test_negative_indention(self) -> None:
        machine = java_description._IndentionMachine()

        machine.write("hello\n")

        value_error = None  # type: Optional[ValueError]

        try:
            machine.dedent()
        except ValueError as a_value_error:
            value_error = a_value_error

        assert value_error is not None
        self.assertEqual("Unexpected negative indention", str(value_error))


if __name__ == "__main__":
    unittest.main()



# NOTE (mristin, 2022-06-19):
# We originally wrote this test to check that the concrete class with descendants
# also implements the interface.


class Parent:
    pass


class Child(Parent):
    pass


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
